#include <iostream>
#include <numeric>
#include "arrivalchain.h"
#include "general.h"
#include "queues.h"



ArrivalChain::ArrivalChain(int duration, int travel_time){
    // create arrival chain that starts with travel_time seconds of state 'p'
    // this is arrival chain for an internal lane
    // duration is used to reserve enough place in the chain
    int length = duration + travel_time + EXTRA_TIME_IN_THE_END_OF_SIMULATION;
    chain.reserve(length);  // more than duration is added as usually arrivals are known for several seconds ahead
    last_fully_known = travel_time - 2; // for what second the full info together with transitions is known
    chain.resize(travel_time);
    for (int i = 0; i < travel_time; ++i){
        // add new 'p' state
        chain[i].emplace('p', 1.);
        if (i != travel_time - 1)
            chain[i]['p'].addTransition('p', 1.); // add transitions if not at second travel_time - 1
    }
}


bool isArrival(int second, int & index, std::vector<int>& arrivals, int & arrival){
    // check if second is an arrival second, if so, update arrival to next arrival or -1 (if there are no arrivals anymore), increment index and return true
    // otherwise return false
    if (second == arrival){
        ++index;
        if (int(arrivals.size()) > index){
            arrival = arrivals[index];
        } else {
            arrival = -1;
        }
        return true;
    } 
    return false;
}


ArrivalChain::ArrivalChain(int duration, int travel_time, std::vector<int>& arrivals, int starting_second){
    // create arrival chain when some arrivals are already known
    // this is for internal lane
    char a_key;
    int length = duration + starting_second + travel_time + EXTRA_TIME_IN_THE_END_OF_SIMULATION;
    chain.reserve(length);
    last_fully_known = travel_time + starting_second - 2; // second for which the full info together with transitions is known
    
    // add arrivals
    int arrival = -1, index = 0;
    int last_second = travel_time + starting_second; // number of seconds for which arrival chain can be already build
    int size = arrivals.size(); // number of arrivals
    
    // update arrival and last second if some arrivals are known
    if (size > 0){
        arrival = arrivals[0]; // first arrival second
        last_second = std::max(last_second, arrivals[size - 1] + 1);
        last_fully_known = last_second - 2; // -1 because indexes start with 0, and -1 more because for the last state the transitions are not known
    }
    
    // create arrival states
    chain.resize(last_second);
    a_key = isArrival(0, index, arrivals, arrival) ? 'a': 'p'; // set a_key to 'a' if there is an arrival and to 'p' (pause) otherwise
    for (int i = 0; i < last_second; ++i){  // add all arrivals also if there are after the travel time
        char next_a_key = isArrival(i + 1, index, arrivals, arrival) ? 'a': 'p'; // find next key
        chain[i].emplace(a_key, 1.);
        if (i != last_second - 1) // add transition if not in the end
            chain[i][a_key].addTransition(next_a_key, 1.);
        a_key = next_a_key;
    }
}


ArrivalChain::ArrivalChain(std::vector<double>& arrival_rates){
    // create arrival chain for an external lane
    int length = arrival_rates.size();
    chain.reserve(length + 1);
    last_fully_known = length + 1; // in this case all the arrivals are known beforehead
    for (int i = 0; i < length; ++i){
        std::map<char, ArrivalState> new_states;
        double lamb = arrival_rates[i];
        
        // add states
        if (lamb > 0.) // add state 'a'
            new_states['a'] = ArrivalState (lamb);
        if (1. > lamb) // add state 'p'
            new_states['p'] = ArrivalState (1. - lamb);
        
        // add transitions
        double new_lamb = 0;
        if (i != length - 1){ 
            new_lamb = arrival_rates[i + 1];
        } 
        if (lamb > 0) // add transitions for state 'a'
            new_states['a'].addTransition('p', lamb);
        if (1. > lamb && new_lamb > 0){ // add transitions from 'p' to 'a' and 'p'
            new_states['p'].addTransition('a', new_lamb);
            if (lamb + new_lamb < 1.)
                new_states['p'].addTransition('p', 1. - lamb - new_lamb);
        } else if (1. > lamb) { // add transitions from 'p' to 'p'
            new_states['p'].addTransition('p', 1. - lamb);
        }
        if (lamb + new_lamb > 1.)
            printWarning("Total arrival rate in consecutive seconds is higher than 1.!");
        chain.push_back(new_states);
    }
    // add last 'p' state
    std::map<char, ArrivalState> new_states;
    new_states['p'] = ArrivalState (1.);
    chain.push_back(new_states);
}


ArrivalChain::ArrivalChain(std::vector<double>& arrival_rates, std::vector<int> & arrivals, int starting_second){
    // create arrival chain for external lane when some arrivals are known
    int arrivals_can_be_from = starting_second + 1; // from when the arrival rates should play a role
    int length = arrival_rates.size();
    chain.reserve(length + 1);
    last_fully_known = length + 1;
    
    // start with known arrivals
    if (arrivals.size() > 0) {
        char a_key;
        int arrival = arrivals[0], index = 0;
        a_key = isArrival(0, index, arrivals, arrival) ? 'a': 'p';
        for (int i = 0; arrival != -1; ++i){  // add all arrivals also if there are after the travel time
            std::map<char, ArrivalState> new_states;
            char next_a_key = isArrival(i + 1, index, arrivals, arrival) ? 'a': 'p';
            new_states[a_key] = ArrivalState (1.);
            new_states[a_key].addTransition(next_a_key, 1.);
            a_key = next_a_key;
            chain.push_back(new_states);
        }
        // add last arrival state
        std::map<char, ArrivalState> new_states;
        a_key = 'a';
        new_states[a_key] = ArrivalState (1.);
        new_states[a_key].addTransition('p', 1.);
        chain.push_back(new_states);
        
        // add pause state after the last arrival state
        a_key = 'p';
        new_states.clear();
        new_states[a_key] = ArrivalState (1.);
        arrivals_can_be_from = chain.size() + 1;
        if (arrivals_can_be_from < length) { // add transitions
            double lamb = arrival_rates[arrivals_can_be_from];
            if (lamb > 0.) // add state 'a'
                new_states[a_key].addTransition('a', lamb);
            if (1. > lamb) // add state 'p'
                new_states[a_key].addTransition('p', 1 - lamb);
        }
        chain.push_back(new_states);
    } else {
        // add 'p' states for the first arrivals_can_be_from seconds
        for (int i = 0; i < arrivals_can_be_from; ++i){
            std::map<char, ArrivalState> new_states;
            new_states['p'] = ArrivalState (1.);
            char a_key = 'p';
            if (i < arrivals_can_be_from - 1) {
                new_states[a_key].addTransition('p', 1);
            } else {
                double lamb = arrival_rates[arrivals_can_be_from];
                if (lamb > 0.) // add tr to state 'a'
                    new_states[a_key].addTransition('a', lamb);
                if (1. > lamb) // add tr to state 'p'
                    new_states[a_key].addTransition('p', 1 - lamb);
            }
            chain.push_back(new_states);
        }
    }
    
    // continue with random arrivals
    for (int i = arrivals_can_be_from; i < length; ++i){
        std::map<char, ArrivalState> new_states;
        double lamb = arrival_rates[i];
        
        // add states
        if (lamb > 0.) // add state 'a'
            new_states['a'] = ArrivalState (lamb);
        if (1. > lamb) // add state 'p'
            new_states['p'] = ArrivalState (1. - lamb);
        
        // add transitions
        double new_lamb = 0; // next arrival rates
        if (i != length - 1){ 
            new_lamb = arrival_rates[i + 1];
        } 
        if (lamb > 0) // add transitions for state 'a'
            new_states['a'].addTransition('p', lamb);
        if (1. > lamb && new_lamb > 0){ // add transitions from 'p' to 'a' and 'p'
            new_states['p'].addTransition('a', new_lamb);
            if (lamb + new_lamb < 1.)
                new_states['p'].addTransition('p', 1. - lamb - new_lamb);
        } else if (1. > lamb) { // add transitions from 'p' to 'p'
            new_states['p'].addTransition('p', 1. - lamb);
        }
        if (lamb + new_lamb > 1.)
            printWarning("Total arrival rate in consecutive seconds is higher than 1.!");
        chain.push_back(new_states);
    }
    // add last 'p' state
    std::map<char, ArrivalState> new_states;
    new_states['p'] = ArrivalState (1.);
    chain.push_back(new_states);
}


void ArrivalChain::printStatesProbabilities (int second){
    // print state keys and probabilities
    std::map<char, ArrivalState> states = chain.at(second);
    std::cout << "The arrival chain for second " << second << ": ";
    for (auto const& x : states){
        std::cout << x.first << ": " << states[x.first].getProbability() << ", ";
    }
    std::cout << "\n";
}


void ArrivalChain::printStates (int second) const{
    // print states together with transitions
    std::map<char, ArrivalState> states = chain.at(second);
    std::cout << "\t\tStates of the arrival chain at second " << second << std::endl;
    for (auto const& x : states){
        std::cout << "\t" << x.first << ": probability = " << states[x.first].getProbability() << ", transition flows (not probabilities):\n";
        states[x.first].printTransitions();
    }
}


void ArrivalChain::addFlow ( int second, char a_key_from, char a_key_to, double flow ){
    // add transition flow from a_key_from to a_key_to at second
    chain[second][a_key_from].addTransition(a_key_to, flow);
    addStateProbability(second + 1, a_key_to, flow); // add state probability for the next state
}


void ArrivalChain::addStateProbability(int second, char a_key, double value){
    // add probability to state a_key at second
    if (int(chain.size()) == second) { // note that it should not happen that state probability is added at some second afterwards
        chain.emplace_back(); // add new second 
    }
    if (isInMap(chain[second], a_key)) 
        chain[second][a_key].addProbability(value); // if such an element exists, just add probability
    else
        chain[second].emplace(a_key, value); // otherwise add state
}


std::map<char, double> ArrivalChain::getNormalisedTransitions(int second, char key){
    // get transitions normalised by the probability of state key at second
    // return an empty map if there is no such state
    if (isInMap(chain[second], key))
        return chain[second][key].getNormalisedTransitions(chain[second][key].getProbability());
    else 
        return std::map<char, double>();
}


std::map<char, std::map<char, double>> ArrivalChain::getNormalisedTransitions(int second){
    // get normalised transitions (transition flows divided by probability of state) for every state at second
    std::map<char, std::map<char, double> > result;
    for (auto & a_from: chain[second]){
        result[a_from.first] = a_from.second.getNormalisedTransitions(a_from.second.getProbability());
    }
    return result;
}


std::map<char, std::map<char, double> > ArrivalChain::getNormalisedTransitionsToKeys(int second){
    // get normalised transitions ordered with key_to, key_from 
    std::map<char, std::map<char, double> > results;
    if (second >= int(chain.size())){ // incorrect second
        std::cout << second << ' ';
        std::cout << "Trying to get normalised transitions for a non-existent second of arrival chain!\n";
        return results;
    }
    // find normalised transitions and store in right indexes
    for (auto const& a_from: chain[second]){
        char a_key_from = a_from.first;
        for (auto const& tr_element: getNormalisedTransitions(second, a_key_from)){
            char a_key_to = tr_element.first;
            double prob = tr_element.second;
            results[a_key_to][a_key_from] = prob;
        }
    }
    return results;
}


std::map<char, std::map<char, double> > ArrivalChain::getNormalisedTransitionsToKeys(int second, int downstream_second, Queues& queues){
    // get normalised transitions when not all transitions are known yet
    // these are transitions for the case of not full queue
    std::map<char, std::map<char, double> > result;
    for (const char & a_key: {'a', 'p'}){ // no 'b' state since the queue is not full
        if (last_fully_known >= second){ // all transitions are known but not all queue probabilities transitioned
            if (hasState(second, a_key)){
                chain[second][a_key].getTransitionsForUntransitioned(result, a_key, queues.getAlmostFullProbability(second, a_key), true);
            }
        } else { // some transitions are known
            if (hasState(second, a_key)){
                chain[second][a_key].getTransitionsForUntransitioned(result, a_key, queues.getAlmostFullProbability(second, a_key), 
                                                                     queues.getTotalUntransitionedQueue(second, downstream_second, a_key));
            }
        }
    }
    return result;
}

std::map<char, std::map<char, double> > ArrivalChain::getNormalisedFullTransitionsToKeys(int second, Queues& queues){
    // get normalised transitions when not all transitions are known yet
    // these are transitions for the case of full queue
    std::map<char, std::map<char, double> > result;
    for (const char & a_key: {'a', 'p', 'b'}){
        if (hasState(second, a_key)){
            chain[second][a_key].getTransitionsForUntransitionedFull(result, a_key, queues.getAlmostFullProbability(second, a_key), last_fully_known >= second);
        }
    }
    return result;
}



std::map<char, ArrivalState> & ArrivalChain::getStates(int second){
    // get states at second
    return chain[second];
}

char ArrivalChain::getFirstState(int second){
    // get the first state of the chain (there should be only one)
    return chain[second].begin()->first;
}


bool ArrivalChain::hasState(int second, char a_key) const{
    // is there a particular state at second 
    return int(chain.size()) > second && isInMap(chain[second], a_key);
}

double ArrivalChain::getStateProbability ( int second, char a_key ) const{
    // get state probability of a_key at second
    if (int(chain.size()) <= second){
        printStates(second - 1);
        throw "Wrong chain size\n";
    } else if (!isInMap(chain[second], a_key)){
        printStates(second);
        throw "No such key\n";
    }
    return chain.at(second).at(a_key).getProbability(); 
}



std::map<KeyIA, double> ArrivalChain::getUnknownTransitionProbabilities(int second){
    std::map<KeyIA, double> result;
    double prob;
    char a_key;
//     std::cout << "Getting untransitioned probabilities " << second << std::endl;
//     std::cout << chain.size() << "\n";
    for (int i = second; i < int(chain.size()); ++i){
//         std::cout << i << "\n";
//         printStates(i);
        for (auto const & iter: chain[i]){
            a_key = iter.first;
            if ((a_key != 'p') || (i == second)){ // if i > second and a_key = p, then prob should be 0 (even if it is not)
                prob = chain[i][a_key].getUnknownTransitionProbability();
//                 std::cout << i << " " << a_key << " -> " << prob << std::endl;
                if (prob > 0)
                    result[KeyIA(i, a_key)] = prob;
            }
        }
    }
    return result;
}

double ArrivalChain::getUnknownTransitionProbability(int second, char a_key){
    return chain[second][a_key].getUnknownTransitionProbability();
}

double ArrivalChain::getUntransitionedProbability(int second, char a_key){
    return chain[second][a_key].getUntransitionedProbability();
}




std::map<KeyIA, double> ArrivalChain::getBlockingProbabilities(int downstream_second, Queues& queues){
    std::map<KeyIA, double> results;
    char a_key;
    double untransitioned_prob;
//     std::cout << "getting blocking probabilities\n";
//     std::cout << "chain size: " << chain.size() << '\n';
//     std::cout << "downstream second: " << downstream_second << "\n";
    for (int i = downstream_second; i < int(chain.size()); ++i){
//         std::cout << i << '\n';
//         switchStyle("yellow");
//         printStates(i);
// // //         queues.printQueues(i);
//         switchStyle("reset");
//         double sum = std::accumulate(chain[i].begin(), chain[i].end(), 0., [](double value, auto & state) 
//                                      { std::cout << state.first << " " << state.second.getProbability() << "\n"; return value + state.second.getProbability();});
//         std::cout << sum << " (" << 1 - sum << ")\n"; 
        for (auto const & iter: chain[i]){
            a_key = iter.first;
//             std::cout << a_key << '\n';
            if (a_key == 'b') {
                results[KeyIA(i, a_key)] = 1;
                continue;
            }
            if ((a_key != 'p') || (i == downstream_second)){ // if i > second and a_key = p, then prob is not important
                //untransitioned_prob = chain[i][a_key].getUntransitionedProbability(); <- incorrect because we need unknown probability!
                untransitioned_prob = chain[i][a_key].getUnknownTransitionProbability();
//                 std::cout << untransitioned_prob << '\n';
                if (untransitioned_prob > almost_zero)
//                     results[KeyIA(i, a_key)] = (queues.getAlmostFullProbability(i, a_key) - getTransitionFlow(i, a_key, 'b')) / untransitioned_prob;
                    results[KeyIA(i, a_key)] = queues.getAlmostFullUntransitionedProbability(i, a_key) / untransitioned_prob;
            }
        }
    }
    return results;
}

bool ArrivalChain::checkTotalProbability(int second){
    double total = 0;
    for (auto const & a_iter: chain[second]){
        double prob = chain[second][a_iter.first].getProbability();
        total += prob;
        if (prob < 0){
            std::cout << second << "\n";
            printStates(second - 1);
            printStates(second);
            std::cout << a_iter.first << " " << prob << "\n";
//             throw "Negative probability of the state\n";
            std::cout << "Negative probability of the state\n";
            return false;
        }
    }
    if (total + eps < 1){
        printStates(second);
        std::cout << total << "\n";
        std::cout << 1 - total << "\n";
//         throw "Too low total probability of arrival chain\n";
        std::cout << "Too low total probability of arrival chain\n";
        return false;
    }
    return true;
}

bool ArrivalChain::checkTotalTransitionProbability(int second){
//     printStates(second);
    bool everything_is_fine = true;
    for (auto const & a_iter: chain[second]){
//         std::cout << second << " " << a_iter.first << " <- checking...\n";
        everything_is_fine = everything_is_fine && chain[second][a_iter.first].checkTransitions(chain[second][a_iter.first].getProbability());
    }
    return everything_is_fine;
}

bool ArrivalChain::checkTransitionedFlows(int second){
    bool everything_is_fine = true;
    for (int i = second; i < int(chain.size()); ++i){
        for (auto & a_iter: chain[i]){
//             std::cout << i << " " << a_iter.first << " <- checking...\n";
            everything_is_fine = everything_is_fine && a_iter.second.checkTransitionedFlows((second == i));
        }    
    }
    return everything_is_fine;
}



void ArrivalChain::addDefaultTransitions(int travel_time_second, double unblocked){
    double prob;
    for (auto & a_key: {'a', 'p', 'b'}){
        if (hasState(travel_time_second, a_key)){
            if (a_key != 'b'){
                prob = chain[travel_time_second][a_key].addDefaultTransition('p');
                addStateProbability(travel_time_second + 1, 'p', prob);
            } else {
                prob = chain[travel_time_second][a_key].addDefaultBTransition(unblocked);
                addStateProbability(travel_time_second + 1, 'b', prob);
                addStateProbability(travel_time_second + 1, 'p', unblocked);
            }
        }
    }
    if (last_fully_known < travel_time_second - 1)
        throw "Added default transitions after a previously unknown second.";
    else 
        last_fully_known = std::max(last_fully_known, travel_time_second);
        
}

void ArrivalChain::addTransitionedProbabilities(int second, MapCharChar & transitioned, MapCharChar & transitioned_full, double unblocked){
    if (unblocked > almost_zero){
        double total_b = 0;
        total_b = std::accumulate(transitioned_full.begin(), transitioned_full.end(), 0., [] (double value, const auto & pair) 
                                  {return value + getValue(pair.second, 'b', 0.);});
//         for (auto & a_iter: transitioned_full){
//             if (isInMap(a_iter.second, 'b')){
//                 total_b += a_iter.second['b'];
//             }
//         }
        // total_b should be >= unlocked
        if (total_b + small_eps < unblocked){
            std::cout << total_b << " " << unblocked << " " << unblocked - total_b << "\n";
            throw "Total blocked is less than unblocked\n";
        }
        unblocked /= total_b;
        unblocked = std::min(unblocked, 1.);
        for (auto & a_iter: transitioned_full){
            if (isInMap(a_iter.second, 'b')){
                double prob = a_iter.second['b'];
                addValue(a_iter.second, 'b', -unblocked * prob);
                addValue(a_iter.second, 'p', unblocked * prob);
            }
        }
    }
//     std::cout << transitioned_full << "<- after\n";
//     switchStyle("yellow");
//     std::cout << second << "\n";
//     std::cout << transitioned << "\n";
//     std::cout << transitioned_full << "\n";
//     std::cout << unblocked << "\n";
//     switchStyle("reset");
    for (const char & a_key: {'a', 'p', 'b'}){
        if (hasState(second, a_key)){
            chain[second][a_key].addTransitionedProbabilities(transitioned[a_key], transitioned_full[a_key]);
        }
    }
}



