#include <algorithm>
#include <cmath>
#include "general.h"
#include "lane.h"
#include "lastarrivals.h"
#include "departurechain.h"
#include "tail.h"

int findShortTravelTime(int capacity){
    return std::ceil((capacity - 1) * bumper_to_bumper * (1 / free_flow_speed + 1 / wave_speed) +  bumper_to_bumper / free_flow_speed) 
           - std::ceil((capacity - 1) * bumper_to_bumper / wave_speed) + 1;
    
}



Lane::Lane(int id, int duration, std::vector<double>& arrival_rates, int cycle, int offset, int red, int green, int capacity, 
           std::map<int, double>& routing_probabilities): // external lane
    short_travel_time(findShortTravelTime(capacity)), arrival_chain(arrival_rates), 
    departure_chain(duration, cycle, offset, red, green),
    queues(duration, capacity, arrival_chain.getStates(0), departure_chain.getFirstState()), capacity(capacity), type('e'), id(id),
    routing_probabilities(routing_probabilities), travel_time(-1), starting_second(0){
    if (short_travel_time <= 1){
        throw "Very short lane!\n";
    }
}


Lane::Lane(int id, int duration, double travel_time, int cycle, int offset, int red, int green, int capacity, 
           std::map<int, double>& routing_probabilities): // internal lane
    short_travel_time(findShortTravelTime(capacity)), arrival_chain(duration, short_travel_time), departure_chain(duration, cycle, offset, red, green),
    queues(duration, capacity, travel_time, short_travel_time, departure_chain.getFirstState()), capacity(capacity), type('i'), id(id),
    routing_probabilities(routing_probabilities), travel_time(travel_time), starting_second(0){
    if (short_travel_time <= 1){
        throw "Very short lane!\n";
    }
}


Lane::Lane(int id, int duration, std::vector<double>& arrival_rates, std::vector<int> & starts, std::vector<int> & ends, int capacity, 
           std::map<int, double>& routing_probabilities): // external lane given schedule
    short_travel_time(findShortTravelTime(capacity)), arrival_chain(arrival_rates), 
    departure_chain(starts.size() > 0 ? DepartureChain(duration, starts,  ends) : DepartureChain(duration)), 
    queues(duration, capacity, arrival_chain.getStates(0), departure_chain.getFirstState()), capacity(capacity), type('e'), id(id),
    routing_probabilities(routing_probabilities), travel_time(-1), starting_second(0){
    if (short_travel_time <= 1){
        throw "Very short lane!\n";
    }
}


Lane::Lane(int id, int duration, double travel_time, std::vector<int> & starts, std::vector<int> & ends, int capacity, 
           std::map<int, double>& routing_probabilities): // internal lane given schedule
    short_travel_time(findShortTravelTime(capacity)), arrival_chain(duration, short_travel_time), 
    departure_chain(starts.size() > 0 ? DepartureChain(duration, starts,  ends) : DepartureChain(duration)),
    queues(duration, capacity, travel_time, short_travel_time, departure_chain.getFirstState()), capacity(capacity), type('i'), id(id),
    routing_probabilities(routing_probabilities), travel_time(travel_time), starting_second(0){
    if (short_travel_time <= 1){
        throw "Very short lane!\n";
    }
}


Lane::Lane(int id, int duration, std::vector<double>& arrival_rates, std::vector<int> & starts, std::vector<int> & ends, int capacity, 
           std::map<int, double>& routing_probabilities, int queue_length, std::vector<Pair> & green_periods_so_far, 
           Pair & queue_state, Pair & lane_state, std::vector<int> & arrivals, int intention, int starting_second): // external lane given schedule + state
    short_travel_time(findShortTravelTime(capacity)), 
    arrival_chain(arrival_rates, arrivals, starting_second), 
    departure_chain(starts.size() > 0 ? DepartureChain(duration, starts, ends, green_periods_so_far, intention, starting_second, 
                                                       lane_state, queue_state, queue_length == 0)
                                      : DepartureChain(duration, green_periods_so_far, intention, starting_second, 
                                                       lane_state, queue_state, queue_length == 0)), 
    queues(duration, capacity, queue_length, queue_state, arrivals.size() > 0 && arrivals[0] == starting_second, starting_second), capacity(capacity), type('e'), id(id),
    routing_probabilities(routing_probabilities), travel_time(-1), starting_second(starting_second){
    if (short_travel_time <= 1){
        throw "Very short lane!\n";
    }
}



Lane::Lane(int id, int duration, double travel_time, std::vector<int> & starts, std::vector<int> & ends, int capacity, 
           std::map<int, double>& routing_probabilities, int queue_length, std::vector<Pair> & green_periods_so_far, 
           Pair & queue_state, Pair & lane_state, std::vector<int> & arrivals, int intention, int starting_second): // internal lane given schedule + state
        short_travel_time(findShortTravelTime(capacity)), 
        arrival_chain(duration, short_travel_time, arrivals, starting_second), 
        departure_chain(starts.size() > 0 ? DepartureChain(duration, starts, ends, green_periods_so_far, intention, starting_second, 
                                                           lane_state, queue_state, queue_length == 0)
                                          : DepartureChain(duration, green_periods_so_far, intention, starting_second, 
                                                           lane_state, queue_state, queue_length == 0)), 
        queues(duration, capacity, travel_time, queue_length, queue_state, arrival_chain, departure_chain, starting_second), 
        capacity(capacity), type('i'), id(id),
        routing_probabilities(routing_probabilities), travel_time(travel_time), starting_second(starting_second){
    if (short_travel_time <= 1){
        throw "Very short lane!\n";
    }
}


void Lane::printRoutingProbabilities(){
    std::cout << "lane " << id << " " << this->routing_probabilities << " <- routing\n";
}


void Lane::initialiseLastArrivalsInTheBeginning(Pair lane_state, std::vector<Lane> & lanes, int starting_second, bool is_empty, bool is_arrival){
    if (lane_state.second >= 0) { // is green  => initialise last arrivals
        for (auto & iter: routing_probabilities) {
            int u = iter.first;
            last_arrivals[u].initialiseInTheBeginning(starting_second, lane_state, lanes[u], is_empty, is_arrival);
        }
    }
}


double Lane::getAverageArrivals(int from, int to){
    double total = 0;
    for (int i = from; i < to; ++i){
        if (arrival_chain.hasState(i, 'a'))
            total += arrival_chain.getStateProbability(i, 'a');
    }
    return total;
}



void Lane::updateQueues(int step){
    if (type == 'e') {
        queues.updateQueues(step, arrival_chain, departure_chain);
    } else { // internal
        queues.updateInternalQueues(step, short_travel_time, arrival_chain, departure_chain);
    }
    assert(queues.checkAndPrintTotalProbability(step + 1, true, false));
}

double Lane::getAverageQueue(int second){
    return queues.getAverageQueue(second);
}


double Lane::getAverageQueueInMeters(int second){
    return queues.getAverageQueueInMeters(second);
}

std::vector<double> Lane::getNormalAverages(int duration){
    return queues.getNormalAverages(duration);
}



void updateEmptied(int step, std::map<KeyDA, double> & just_emptied, DepartureChain & departure_chain){
    Pair d_key;
    KeyDA da_key;
    switchStyle("blue");
    for (auto da_iter = just_emptied.begin(); da_iter != just_emptied.end(); /*no increment*/){
        da_key = da_iter->first;
        d_key = da_key.first;
        if (departure_chain.getLastSecond(d_key) < step){
            do {
                d_key = departure_chain.getDefaultZeroTransition(d_key);
            } while (departure_chain.getLastSecond(d_key) < step);
            // was bad d_key
            da_key.first = d_key;
            addValue(just_emptied, da_key, da_iter->second);
            da_iter = just_emptied.erase(da_iter);
        } else {
            // normal d_key -> proceed
            ++da_iter;
        }
    }
    switchStyle("reset");
}


std::map<Pair, double> sumKeyDAMap(std::map<KeyDA, double> my_map){
    std::map<Pair, double> result;
    for (auto & da_iter: my_map){
        addValue(result, da_iter.first.first, da_iter.second);
    }
    return result;
}

void Lane::preprocessLastArrivalsAndDepartureChain(int step, std::vector<Lane>& lanes, std::map<int, double> & travel_times){
    if (departure_chain.isExtendedGreen(step)){
        int u;
        // starting second is initialised in another place
        if (departure_chain.isGreenBegin(step + extra_green) || step == 0) {// initialise in the beginning of a green period
            std::map<KeyDA, double> zero_probs = getZeroProbabilities(step);
            double minus_first_probability = 0;
            for(const auto& iter: routing_probabilities){
                u = iter.first;
                minus_first_probability = last_arrivals[u].initialise(step, zero_probs, lanes[u]);
            }
            departure_chain.setFirstProbability(step, zero_probs, minus_first_probability);
        } else if (step > starting_second){// update otherwise
            std::map<char, std::map<char, double> > ar_transitions = queues.getLastArrivalTransitions(); 
    
            std::map<KeyDA, double> just_emptied_probs = queues.getJustEmptiedProbabilities(step, departure_chain.isSplitLane(), (type == 'i')); // (step, departure_chain, ar_transitions); // TODO: make right

            departure_chain.updateStateProbabilityFromEmptied(step, just_emptied_probs); 
            assert(queues.checkNonZeroProbs(step, departure_chain));
            updateEmptied(step, just_emptied_probs, departure_chain);
            for(const auto& iter: routing_probabilities){
                u = iter.first;
                last_arrivals[u].updateNewSecond(step, departure_chain, ar_transitions, just_emptied_probs, lanes[u], travel_times[u]);
            }
        }
        std::map<KeyDA, double> zero_probs = getZeroProbabilities(step); // important to get zero probs after (!) getting just_emptied_probs
        for(const auto& iter: routing_probabilities){
            u = iter.first;
            assert(last_arrivals[u].checkTotalProbabilities(zero_probs, departure_chain));
        }
    }
}



void Lane::processDeparturesOneLane(int step, std::vector<Lane>& lanes, int u, double gamma, std::vector<Tail> & tails, 
                                    std::vector<double> & unblocked_probabilities_of_tails, double travel_time){
    std::map<KeyDA, std::map<Pair, double>> tr_flows_zero;
    std::map<Pair, std::map<Pair, double>> tr_flows_plus; 
    std::map<Pair, std::map<int, double>> unblocking_d_keys_probabilities; 
    LastArrivals new_last_arrivals; 
    std::map<KeyDI, double> unblocking_probabilities;
    if (departure_chain.isExtendedGreen(step)){
        std::map<KeyIA, double> blocking_probabilities = lanes[u].getBlockingProbabilities(step);
        generateTails(step, gamma, u, tails, blocking_probabilities, lanes[u], new_last_arrivals, tr_flows_zero, tr_flows_plus, unblocking_d_keys_probabilities, 
                      unblocking_probabilities, unblocked_probabilities_of_tails, travel_time);
        // update last arrivals
        updateLastArrivalsAfterDepartureToU(step, u, tr_flows_zero, tr_flows_plus, new_last_arrivals, unblocking_probabilities);
        // update departure states (what was blocked)
        departure_chain.updateStates(tr_flows_plus, unblocking_d_keys_probabilities);
        // update departure states in queues
        queues.updateDepartureState(step, tr_flows_zero); // updating R_step to \dot R_step
    }
}


void Lane::processDepartures(int step, std::vector<Lane>& lanes){
    std::map<int, std::map<KeyDA, std::map<Pair, double>>> tr_flows_zero;
    std::map<int, std::map<Pair, std::map<Pair, double>>> tr_flows_plus;
    std::map<Pair, std::map<int, double>> unblocking_d_keys_probabilities;
    std::map<int, LastArrivals> new_last_arrivals;
    std::map<int, std::map<KeyDI, double>> unblocking_probabilities;
    std::vector<double> unblocked_probabilities_of_tails;
    std::vector<Tail> tails;
    if (departure_chain.isExtendedGreen(step)){
        for(const auto& iter: routing_probabilities){
            int u = iter.first;
            double gamma = iter.second;
            std::map<KeyIA, double> blocking_probabilities = lanes[u].getBlockingProbabilities(step);
            generateTails(step, gamma, u, tails, blocking_probabilities, lanes[u], new_last_arrivals[u], tr_flows_zero[u], tr_flows_plus[u], unblocking_d_keys_probabilities, 
                          unblocking_probabilities[u], unblocked_probabilities_of_tails, lanes[u].getFullTravelTime());
            
        }
        updateAfterDepartures(step, tr_flows_zero, tr_flows_plus, new_last_arrivals, unblocking_probabilities, unblocking_d_keys_probabilities);

    }
}

void Lane::updateAfterDepartures(int step,  std::map<int, std::map<KeyDA, std::map<Pair, double>>> & tr_flows_zero, 
                                 std::map<int, std::map<Pair, std::map<Pair, double>>> & tr_flows_plus, std::map<int, LastArrivals> & new_last_arrivals, 
                                 const std::map<int, std::map<KeyDI, double>> & unblocking_probabilities,
                                 const std::map<Pair, std::map<int, double>> & unblocking_d_keys_probabilities){
    std::map<KeyDA, std::map<Pair, double>> tr_flows_zero_sum;
    std::map<Pair, std::map<Pair, double>> tr_flows_plus_sum;
    for(const auto& iter: routing_probabilities){
        int u = iter.first;
        addMap(tr_flows_plus_sum, tr_flows_plus[u]);
        addMap(tr_flows_zero_sum, tr_flows_zero[u]);
    }
    if (routing_probabilities.size() > 0) {
        updateLastArrivals(step, tr_flows_zero, tr_flows_plus, new_last_arrivals, unblocking_probabilities);
        departure_chain.updateStates(tr_flows_plus_sum, unblocking_d_keys_probabilities);
        queues.updateDepartureState(step, tr_flows_zero_sum); // updating R_step to \dot R_step
    }
}


void Lane::updateLastArrivalsAfterDepartureFromOtherLane(int step, int u, std::vector<Tail>& tails, std::map<KeyIA, std::vector<int>> & ordered_tails, 
                                                         std::vector<double> & unblocked_probabilities_of_tails){
    if (departure_chain.isExtendedGreen(step)){ // if green
        if (isInMap(last_arrivals, u)) { // if routing probability is not zero
            last_arrivals[u].updateFromTails(ordered_tails, tails, unblocked_probabilities_of_tails);
        }
    }
}

void multiplyLAByCoef(std::map<KeyIA, double> & la_probs, double coef){
    for (auto & ia_iter: la_probs){
        ia_iter.second *= coef;
    }
}

void addLAToBD(std::map<KeyIA, double> & la_before_departure_d, std::map<KeyIA, double> other_la, double coef){
    for (const auto & ia_iter: other_la){
        addValue(la_before_departure_d, ia_iter.first, ia_iter.second * coef);   
    }
}

void Lane::updateLastArrivalsAfterDepartureToU(int step, int u, const std::map<KeyDA, std::map<Pair, double> >& tr_flows_zero, 
                                               const std::map<Pair, std::map<Pair, double> >& tr_flows_plus, 
                                               LastArrivals& new_last_arrivals, const std::map<KeyDI, double>& unblocking_probabilities){
    double gamma,prob;
    Pair d_key, new_d_key;
    
    std::map<KeyDA, double> zero_probs  = queues.getZeroProbabilities(step);
    
    // first update the last-arrival probabilities for downstream lane u
    gamma = routing_probabilities.at(u);
    for (auto & da_iter: last_arrivals[u].last_arrivals){
        if (da_iter.first.second == 'a') {
            addMap(new_last_arrivals.last_arrivals[da_iter.first], last_arrivals[u].last_arrivals[da_iter.first], 1 - gamma);
        } else {
            addMap(new_last_arrivals.last_arrivals[da_iter.first], last_arrivals[u].last_arrivals[da_iter.first]);
        }
    }
    for (auto & da_iter: last_arrivals[u].last_delayed_arrivals){
        if (da_iter.first.second == 'a') {
            addMap(new_last_arrivals.last_delayed_arrivals[da_iter.first], last_arrivals[u].last_delayed_arrivals[da_iter.first], 1 - gamma);
        } else {
            addMap(new_last_arrivals.last_delayed_arrivals[da_iter.first], last_arrivals[u].last_delayed_arrivals[da_iter.first]);
        }
    }
    for (auto & d_iter: last_arrivals[u].la_before_departure){
        d_key = d_iter.first;
        if (departure_chain.isDepartureSecond(step, d_key)){
            double gamma_d = departure_chain.getRoutingProbabilities(d_key, routing_probabilities, u);
            double dep_prob = departure_chain.getDepartureProbability(d_key);
            addMap(new_last_arrivals.la_before_departure[d_key], last_arrivals[u].la_before_departure[d_key], 1 - gamma_d * dep_prob);
        } else if (departure_chain.getLastSecond(d_key) >= step){
            addMap(new_last_arrivals.la_before_departure[d_key], last_arrivals[u].la_before_departure[d_key]);
        }
    }
    for (auto & di_iter: unblocking_probabilities){
        d_key = di_iter.first.first;
        int i = di_iter.first.second;
        prob = di_iter.second;
        double not_more = new_last_arrivals.la_before_departure[d_key][KeyIA(i, 'b')];
        if (prob > not_more + small_eps * not_more){
            throw "Incorrect unblocking probability, it is higher than probability that the queue was blocked!\n";
        }
        prob = std::min(prob, not_more);
        addValue(new_last_arrivals.la_before_departure, d_key, KeyIA(i, 'p'), prob);
        addValue(new_last_arrivals.la_before_departure, d_key, KeyIA(i, 'b'), -prob);
    }
    last_arrivals[u] = new_last_arrivals;
    // Second, change PLA, PLDA, PBD to PBD for blocked states for other lanes if the departure was blocked
    for(const auto& iter: routing_probabilities){
        int u_other = iter.first;
        if (u_other == u)
            continue;
        for (auto & da_iter: tr_flows_zero){ // how much probability was blocked
            KeyDA da_key = da_iter.first;
            double prob_zero = getValue(zero_probs, da_key, 0.);
            if (prob_zero > almost_zero){
                double sum_of_probs = 0;
                for (auto & d_iter: da_iter.second){
                    new_d_key = d_iter.first;
                    prob = d_iter.second;
                    sum_of_probs += prob;
                    addLAToBD(last_arrivals[u_other].la_before_departure[new_d_key], last_arrivals[u_other].last_arrivals[da_key], prob / prob_zero);
                    addLAToBD(last_arrivals[u_other].la_before_departure[new_d_key], 
                              sumNestedMaps(last_arrivals[u_other].last_delayed_arrivals[da_key]), prob / prob_zero);
                }
                if (1 - sum_of_probs / prob_zero > almost_zero){ // multiply by 1 - sum_of_probs / prob_zero
                    multiplyLAByCoef(last_arrivals[u_other].last_arrivals[da_key], 1 - sum_of_probs / prob_zero);
                    for (auto & i_iter: last_arrivals[u_other].last_delayed_arrivals[da_key]){
                        multiplyLAByCoef(i_iter.second, 1 - sum_of_probs / prob_zero);
                    }
                } else {
                    last_arrivals[u_other].last_arrivals.erase(da_key);
                    last_arrivals[u_other].last_delayed_arrivals.erase(da_key);
                }
            }
        }
        for (auto & d_iter: tr_flows_plus){
            d_key = d_iter.first;
            double prob_plus = sumMapValues(last_arrivals[u_other].la_before_departure[d_key]);
            if (prob_plus > almost_zero){
                double sum_of_probs = 0;
                for (auto & new_d_iter: d_iter.second){
                    new_d_key = new_d_iter.first;
                    prob = new_d_iter.second;
                    sum_of_probs += prob;
                    addLAToBD(last_arrivals[u_other].la_before_departure[new_d_key], last_arrivals[u_other].la_before_departure[d_key], prob / prob_plus);
                }
                if (1 - sum_of_probs / prob_plus > almost_zero){
                    multiplyLAByCoef(last_arrivals[u_other].la_before_departure[d_key], 1 - sum_of_probs / prob_plus);
                } else {
                    last_arrivals[u_other].la_before_departure.erase(d_key);
                }
            }
        }
    }
    
}

void Lane::updateLastArrivalsAfterAllDepartures(int step){
    if (departure_chain.isExtendedGreen(step)){ // if green
        int u;
        Pair d_key, new_d_key;
//         assert(checkDifferentLastArrivals(last_arrivals)); // may be transferred for one lane but not another
        for(const auto& iter: routing_probabilities){
            u = iter.first;
            for (auto d_iter = last_arrivals[u].la_before_departure.begin(); d_iter != last_arrivals[u].la_before_departure.end(); /*no increment*/){
                d_key = d_iter->first;
                if (departure_chain.isDepartureSecond(step, d_key)){
                    new_d_key = departure_chain.getDefaultPlusTransition(d_key);
                    addMap(last_arrivals[u].la_before_departure[new_d_key], last_arrivals[u].la_before_departure[d_key]);
                    d_iter = last_arrivals[u].la_before_departure.erase(d_iter);
                } else {
                    ++d_iter;
                }
            }
        }
        if (routing_probabilities.size() == 0) {
            departure_chain.updateStateProbabilityAfterDepartureSecond(step);
        } else {
            u = routing_probabilities.begin()->first;
            std::vector<Pair> d_key_to_update;
            for (auto & d_iter: last_arrivals[u].la_before_departure){
                d_key = d_iter.first;
                if (departure_chain.wasDepartureSecond(step, d_key))
                    d_key_to_update.push_back(d_key);
            }
            departure_chain.updateStateProbabilityAfterDepartureSecond(d_key_to_update);
        }
        assert(checkDifferentLastArrivals(last_arrivals));
    }
}



void Lane::updateLastArrivals(int step, const std::map<int, std::map<KeyDA, std::map<Pair, double>>>& tr_flows_zero,
                              const std::map<int, std::map<Pair, std::map<Pair, double>>>& tr_flows_plus, std::map<int, LastArrivals>& new_last_arrivals, 
                              const std::map<int, std::map<KeyDI, double>> & unblocking_probabilities){
    int u, u_other;
    Pair d_key, new_d_key;
    // \sum_{u' \neq u} \phi^{a,dd'}_{0, u'},  \phi^{a,dd'}_{0, u'} = P(R_{i+1, u'} = d'| X_{i,u'} = (0, a, d)), d' is unblocking state:
    std::map<int, std::map<KeyDA, std::map<Pair, double>>> tr_probs_sums_zero;  
    std::map<int, std::map<Pair, std::map<Pair, double>>> tr_probs_sums_plus; // \sum_{u' \neq u} \phi^{dd'}_{+, u'}

    
    std::map<KeyDA, double> zero_probs  = queues.getZeroProbabilities(step);
    for(const auto& iter: routing_probabilities){
        u = iter.first;
        for (auto & da_iter: tr_flows_zero.at(u)){
            double prob_zero = zero_probs[da_iter.first];
            if (prob_zero > almost_zero) {
                for(const auto& iter2: routing_probabilities){
                    u_other = iter2.first;
                    if (u_other != u)
                        addMap(tr_probs_sums_zero[u_other][da_iter.first], da_iter.second, 1 / prob_zero);
                }
            }
        }
        for (auto & d_iter: tr_flows_plus.at(u)){
            d_key = d_iter.first;
            double prob_plus = departure_chain.getStateProbability(d_key);
            if (prob_plus > almost_zero) {
                for(const auto& iter2: routing_probabilities){
                    u_other = iter2.first;
                    if (u_other != u)
                        addMap(tr_probs_sums_plus[u_other][d_key], d_iter.second, 1 / prob_plus);
                } 
            }
        }
    }
    
    for(const auto& iter: routing_probabilities){
        u = iter.first;
        double gamma = iter.second;
        for (auto & da_iter: last_arrivals[u].last_arrivals){
            double sum = sumMapValues(tr_probs_sums_zero[u][da_iter.first]);
            if (da_iter.first.second == 'a') {
                addMap(new_last_arrivals[u].last_arrivals[da_iter.first], last_arrivals[u].last_arrivals[da_iter.first], 1 - gamma - sum);
            } else {
                addMap(new_last_arrivals[u].last_arrivals[da_iter.first], last_arrivals[u].last_arrivals[da_iter.first]);
            }
            for (auto & d_iter: tr_probs_sums_zero[u][da_iter.first]){
                new_d_key = d_iter.first;
                addLAToBD(new_last_arrivals[u].la_before_departure[new_d_key], last_arrivals[u].last_arrivals[da_iter.first], d_iter.second);
            }
        }
        for (auto & da_iter: last_arrivals[u].last_delayed_arrivals){
            double sum = sumMapValues(tr_probs_sums_zero[u][da_iter.first]);
            if (da_iter.first.second == 'a') {
                addMap(new_last_arrivals[u].last_delayed_arrivals[da_iter.first], last_arrivals[u].last_delayed_arrivals[da_iter.first], 1 - gamma - sum);
            } else {
                addMap(new_last_arrivals[u].last_delayed_arrivals[da_iter.first], last_arrivals[u].last_delayed_arrivals[da_iter.first]);
            }
            for (auto & d_iter: tr_probs_sums_zero[u][da_iter.first]){
                new_d_key = d_iter.first;
                addLAToBD(new_last_arrivals[u].la_before_departure[new_d_key], 
                        sumNestedMaps(last_arrivals[u].last_delayed_arrivals[da_iter.first]), d_iter.second);
            }
        }
        for (auto & d_iter: last_arrivals[u].la_before_departure){
            d_key = d_iter.first;
            if (departure_chain.isDepartureSecond(step, d_key)){
                double sum = sumMapValues(tr_probs_sums_plus[u][d_key]); //this is blocking probability due to other downstream lanes
                new_d_key = departure_chain.getDefaultPlusTransition(d_key);
                double gamma_d = departure_chain.getRoutingProbabilities(d_key, routing_probabilities, u);
                double dep_prob = departure_chain.getDepartureProbability(d_key);
                addMap(new_last_arrivals[u].la_before_departure[new_d_key], d_iter.second, 1 - gamma_d * dep_prob - sum);
                for (auto & new_d_iter: tr_probs_sums_plus[u][d_key]){
                    new_d_key = new_d_iter.first;
                    addLAToBD(new_last_arrivals[u].la_before_departure[new_d_key], d_iter.second, new_d_iter.second);
                }
            } else if (departure_chain.getLastSecond(d_key) > step){
                addMap(new_last_arrivals[u].la_before_departure[d_key], d_iter.second);
            }
        }
        for (auto & di_iter: unblocking_probabilities.at(u)){
            d_key = di_iter.first.first;
            int i = di_iter.first.second;
            double prob = di_iter.second;
            double not_more = new_last_arrivals[u].la_before_departure[d_key][KeyIA(i, 'b')];
            if (prob > not_more + small_eps * not_more){
                throw "Incorrect unblocking probability, it is higher than probability that the queue was blocked!\n";
            }
            prob = std::min(prob, not_more);
            addValue(new_last_arrivals[u].la_before_departure, d_key, KeyIA(i, 'p'), prob);
            addValue(new_last_arrivals[u].la_before_departure, d_key, KeyIA(i, 'b'), -prob);
        }
    }
    last_arrivals = new_last_arrivals;
}



std::map<KeyIA, double> Lane::getUnknownTransitionsProbabilities(int second){
    return arrival_chain.getUnknownTransitionProbabilities(second);
}

std::map<KeyDA, double> Lane::getZeroProbabilities(int second){
    return queues.getZeroProbabilities(second);
}

std::map<char, double> Lane::getBTransitions(int second){
    if (second == 652 && id == 92)
        queues.printQueues(639);
    return arrival_chain.getNormalisedTransitions(second, 'b');
}


std::map<char, std::map<char, double> > Lane::getAllArTransitions(int second){
    return queues.getAllArTransitions(second, arrival_chain);
}



std::map<KeyIA, double> Lane::getBlockingProbabilities(int step){
    int downstream_second = step + short_travel_time - 1; //this is downstream lane
    return arrival_chain.getBlockingProbabilities(downstream_second, queues);
}


template <class key_type>
void addBlockingTail(double blocking_prob, std::vector<Tail>& tails, int i, char a_key, std::map<Pair, 
                     std::map<int, double>> & unblocking_d_keys_probabilities, int u, std::map<Pair, double> & to_unblocking_d_keys, 
                     key_type da_key, std::map<key_type, std::map<Pair, double>> & tr_flows, std::map<Pair, MapIAProb> & la_before_departure){
    if (blocking_prob > almost_zero) {
        bool is_b = (a_key == 'b');
        if (!is_b)
            tails.emplace_back(i, a_key, i + 1, 'b', blocking_prob); // blocked  
        for (auto & d_iter: to_unblocking_d_keys){
            if (d_iter.second > almost_zero){
                Pair unblock_d_key = d_iter.first;
                double prob = d_iter.second * blocking_prob;
                addValue(tr_flows, da_key, unblock_d_key, prob);
                addValue(unblocking_d_keys_probabilities, unblock_d_key, u, prob);
                addValue(la_before_departure, unblock_d_key, KeyIA((is_b) ? i: i + 1, 'b'), prob);
            }
        }
    }
}

template <class type>
inline void addSimpleTail(double probability, std::vector<Tail>& tails, int i, char a_key, 
                          int i_to, std::map<type, MapIAProb> & some_last_arrivals, type key){
    tails.emplace_back(i, a_key, i_to, 'a', probability); // unblocked
    addValue(some_last_arrivals, key, KeyIA(i_to, 'a'), probability);
}

template <class type>
void addTail(double non_blocking_prob, std::vector<Tail>& tails, int i, char a_key, double i_to_double, 
             std::map<type, MapIAProb> & some_last_arrivals, type key){
    int i_to = std::floor(i_to_double);
    double fractional_part = i_to_double - i_to;
    if (fractional_part > 0){
        addSimpleTail(non_blocking_prob * fractional_part, tails, i, a_key, i_to + 1, some_last_arrivals, key);
    }
    addSimpleTail(non_blocking_prob * (1 - fractional_part), tails, i, a_key, i_to, some_last_arrivals, key);
}

void addDelayedTail(double non_blocking_prob, std::vector<Tail>& tails, int i, char a_key, double i_to_double,
             std::map<KeyDA, std::map<int, MapIAProb>> & last_delayed_arrivals, KeyDA da_key){
    int i_to = std::floor(i_to_double);
    double fractional_part = i_to_double - i_to;
    if (fractional_part > 0){
        addSimpleTail(non_blocking_prob * fractional_part, tails, i, a_key, i_to + 1, last_delayed_arrivals[da_key], i_to + 3);
    }
    addSimpleTail(non_blocking_prob * (1 - fractional_part), tails, i, a_key, i_to, last_delayed_arrivals[da_key], i_to + 2);
}


std::map<KeyIA, std::map<Pair, double>> getIAToUnblockDKeys(int min_second, const std::map<KeyIA, std::map<Pair, double>> & almost_full_d_keys, 
                                                            DepartureChain & upstream_departure_chain,
                                                            DepartureChain & downstream_departure_chain, int capacity){
    Pair d_key, unblock_d_key;
    std::map<KeyIA, std::map<Pair, double>> result;
    int new_second, total_reaction_time; // = std::ceil(std::ceil((capacity - 1) * bumper_to_bumper * (1 / wave_speed + 1 / free_flow_speed) + 
                                         //                     bumper_to_bumper / free_flow_speed) - (capacity - 1) * bumper_to_bumper / free_flow_speed)
    total_reaction_time = std::ceil((capacity - 1) * bumper_to_bumper /  wave_speed);
    for (auto & ia_iter: almost_full_d_keys){
        for (auto & d_iter: ia_iter.second){
            d_key = d_iter.first;
            new_second = upstream_departure_chain.getFirstNextGreenSecond(std::max(min_second, 
                                                                                   downstream_departure_chain.getNextUnblock(d_key) + total_reaction_time));
            unblock_d_key = Pair(new_second, 0);
            addValue(result, ia_iter.first, unblock_d_key, d_iter.second);
        }
    }
    return result;
}


void setUnblockingProbabilities(const std::map<KeyIA, double> & unblocked, std::map<KeyDI, double> & unblocking_probabilities, 
                                const std::map<KeyIA, std::map<Pair, double>> & ia_to_unblocking_d_keys){
    for (auto & ia_iter: unblocked){
        int i = ia_iter.first.first;
        if (ia_iter.second > almost_zero){
            for (auto & d_iter: ia_to_unblocking_d_keys.at(ia_iter.first)){
                Pair d_key = d_iter.first;
                if (d_iter.second > almost_zero)
                    addValue(unblocking_probabilities, KeyDI(d_key, i + 1), ia_iter.second * d_iter.second); // i+ 1  to point to the end of a blocking tail
            }
        }
    }
}

double getExtraDelay(int step, DepartureChain & departure_chain){
    double extra = 0;
    int seconds_to_green = departure_chain.getExtraSecondsToGreen(step); // this allows negative green seconds
    if (seconds_to_green > 0) {
        // in this case extra delay
        extra = (seconds_to_green * decel + free_flow_speed / 2);
        extra *= extra;  // extra ^ 2
        extra *= (1 / accel + 1/ decel) / (2 * free_flow_speed);
    }
    return extra;
}


double getArrivalSecond(int i_last, char a_key, double new_i){
    return (a_key == 'a') ? std::max(new_i, double(i_last + 2)) : std::max(new_i, double(i_last + 1));
}



template <class type, class type_la>
void generateTailsFromLastArrivals(const std::map<KeyIA, double> & last_arrivals, double gamma, int u, const std::map<KeyIA, double> & blocking_probabilities, 
                                   std::vector<Tail> & tails, double free_flow_arrival_second, 
                                   std::map<type, type_la> & new_last_arrivals_unblocked_tails, type key1, 
                                   std::map<Pair, MapIAProb> & new_last_arrivals_before_departure, type key2,
                                   std::map<type, std::map<Pair, double>>& tr_flows, 
                                   std::map<Pair, std::map<int, double>>& unblocking_d_keys_probabilities,
                                   std::map<KeyIA, std::map<Pair, double>> & ia_to_unblocking_d_keys, 
                                   void (*tail_func)(double, std::vector<Tail>&, int, char, double, std::map<type, type_la> &, type)){
    for (auto & ia_iter: last_arrivals){
        // if very low probability or low probability and no blocking probability available, then skip
        if (ia_iter.second < almost_zero || (ia_iter.second < small_eps && !isInMap(blocking_probabilities, ia_iter.first)))
            continue;
        KeyIA ia_key = ia_iter.first;
        int i = ia_key.first;
        char a_key = ia_key.second;
        // not blocked tails
        double key_blocked_prob = blocking_probabilities.at(ia_key);
        double non_blocking_prob =  gamma * ia_iter.second * (1 - key_blocked_prob);
        if (non_blocking_prob > almost_zero){
            double i_to_double = getArrivalSecond(i, a_key, free_flow_arrival_second);
            tail_func(non_blocking_prob, tails, i, a_key, i_to_double, new_last_arrivals_unblocked_tails, key1);
        }
        // blocked tail
        double blocking_prob = gamma * ia_iter.second * key_blocked_prob;
        addBlockingTail(blocking_prob, tails, i, a_key, unblocking_d_keys_probabilities, u, ia_to_unblocking_d_keys[ia_key], 
                        key2, tr_flows, new_last_arrivals_before_departure);
        
    }
} 

void Lane::generateTails(int step, double gamma, int u, std::vector<Tail> & tails, const std::map<KeyIA, double>& blocking_probabilities, Lane & downstream_lane, 
                         LastArrivals & new_last_arrivals,
                         std::map<KeyDA, std::map<Pair, double>>& tr_flows_zero, std::map<Pair, std::map<Pair, double>>& tr_flows_plus, 
                         std::map<Pair, std::map<int, double>>& unblocking_d_keys_probabilities, std::map<KeyDI, double> & unblocking_probabilities,
                         std::vector<double> & unblocked_probabilities_of_tails, double travel_time){
    // generate free-departing tails
    KeyDA da_key;
    std::map<KeyIA, std::map<Pair, double>> almost_full_d_keys = downstream_lane.getAlmostFullDKeys(step);
    int first_green_second = departure_chain.getFirstNextGreenSecond(step);
    std::map<KeyIA, std::map<Pair, double>> ia_to_unblocking_d_keys = getIAToUnblockDKeys(first_green_second + 1, almost_full_d_keys, departure_chain, 
                                                                                          downstream_lane.departure_chain, downstream_lane.getCapacity());
    
    double extra_delay = getExtraDelay(step, departure_chain);    
    // process free departures
    for (auto & da_iter: last_arrivals[u].last_arrivals){
        if (da_iter.first.second != 'a') // no free departure
            continue;
        da_key = da_iter.first;
        generateTailsFromLastArrivals(da_iter.second, gamma, u, blocking_probabilities, tails, step + travel_time + extra_delay, 
                                      new_last_arrivals.last_arrivals, da_key, new_last_arrivals.la_before_departure, da_key, tr_flows_zero, 
                                      unblocking_d_keys_probabilities, ia_to_unblocking_d_keys, &addTail);
    }
    // process free departures after delayed departures
    for (auto & da_iter: last_arrivals[u].last_delayed_arrivals){
        if (da_iter.first.second != 'a') // no free departure
            continue;
        da_key = da_iter.first;
        for (auto & iter: last_arrivals[u].last_delayed_arrivals[da_key]){
            int i_new = iter.first;
            generateTailsFromLastArrivals(iter.second, gamma, u, blocking_probabilities, tails, std::max(step + travel_time + extra_delay, double(i_new)), 
                                      new_last_arrivals.last_delayed_arrivals, da_key, new_last_arrivals.la_before_departure, da_key, tr_flows_zero, 
                                      unblocking_d_keys_probabilities, ia_to_unblocking_d_keys, &addDelayedTail);
        }
    }
    // generate delayed tails if possible
    for (auto & d_iter: last_arrivals[u].la_before_departure){
        Pair d_key = d_iter.first;
        if (!departure_chain.isDepartureSecond(step, d_key))
            continue;
        // is departure second
        double gamma_d = departure_chain.getRoutingProbabilities(d_key, routing_probabilities, u);
        double departure_probability = departure_chain.getDepartureProbability(d_key);
        if (departure_probability > 0) { // should be so if the sumo_capacities are not strange
            int dep_index = d_key.second;
            double i_final = step + getArrivalMoment(dep_index) - getDepartureMoment(dep_index) + travel_time;  // s + a_k - d_k + t
            generateTailsFromLastArrivals(d_iter.second, gamma_d * departure_probability, u, blocking_probabilities, tails, i_final, 
                                          new_last_arrivals.la_before_departure, departure_chain.getDefaultPlusTransition(d_key), 
                                          new_last_arrivals.la_before_departure, d_key, tr_flows_plus, 
                                          unblocking_d_keys_probabilities, ia_to_unblocking_d_keys, &addTail);
        }
    }
    setUnblockingProbabilities(downstream_lane.processTails(step, tails, unblocked_probabilities_of_tails), unblocking_probabilities, ia_to_unblocking_d_keys);
}


std::map<KeyIA, double> Lane::processTails(int step, std::vector<Tail> & tails, std::vector<double>& unblocked_probabilities_of_tails){
    std::map<KeyIA, double> unblocked; // i: \sum_{T = (u, i-1, a, i, 'b', p)} \pi^T
    
    std::sort(tails.begin(), tails.end(), [](const auto &left, const auto &right) { return left.second_to > right.second_to; });
    unblocked_probabilities_of_tails.reserve(tails.size());
    for (auto & tail: tails){
        double result = queues.processTail(step, tail, arrival_chain, departure_chain);
        unblocked_probabilities_of_tails.push_back(result);
        if (tail.a_key_to == 'b') {
            int i = tail.second_from;
            char a_key = tail.a_key_from;
            addValue(unblocked, KeyIA(i, a_key), result);
        }
    }
    return unblocked;
}

std::map<KeyIA, std::map<Pair, double> > Lane::getAlmostFullDKeys(int second){
    return queues.getAlmostFullDKeys(second + short_travel_time - 1);
}

void Lane::cleanUnprobableStates(int step){
    // remove departure states that have very low probability
    std::vector<Pair> d_keys_to_remove = queues.cleanUnprobableStates(step, departure_chain);
    for(const auto& iter: routing_probabilities){
        int u = iter.first;
        last_arrivals[u].removeStates(d_keys_to_remove);
    }
}

int Lane::getTotalNumberOfQueues(){
    return queues.getTotalNumberOfQueues();
}

void Lane::printDistributionsToFile(int duration, const std::string & file_name){
    printStringToFile("[", file_name);
    std::vector<double> distribution;
    for (int i = 0; i < duration; ++i){
        distribution = queues.getDistribution(i);
        printVectorToFile(distribution, file_name);
        if (i < duration - 1)
            printStringToFile(", ", file_name);
    }
    printStringToFile("]", file_name);
}

double Lane::getAverageAndRemoveQueues(int second){
    return queues.getAverageAndRemove(second);
}


double Lane::getAverageSquareAndRemoveQueues(int second, int start_from /* = 0*/){
    return queues.getAverageSquareAndRemove(second, start_from);
}
