#ifndef LANE_H
#define LANE_H
#include <vector>
#include <map>
#include "queues.h"
#include "arrivalchain.h"
#include "departurechain.h"
#include "tail.h"
#include "lastarrivals.h"

//using Pair = std::pair<int, int>;

class Lane{
    int short_travel_time;
    ArrivalChain arrival_chain;
    DepartureChain departure_chain;
    Queues queues;
    int capacity;
    char type;  // 'i' for internal and 'e' for external
    int id;
    std::map<int, double> routing_probabilities;
    std::map<int, LastArrivals> last_arrivals;
    double travel_time;
    int starting_second;
public:
    Lane(int id, int duration, std::vector<double>& arrival_rates, int cycle, int offset, int red, int green, int capacity, 
         std::map<int, double>& routing_probabilities); // for external lanes fixed cycle
    Lane(int id, int duration, double travel_time, int cycle, int offset, int red, int green, int capacity, 
         std::map<int, double>& routing_probabilities); // for internal lanes fixed cycle
    
    Lane(int id, int duration, std::vector<double>& arrival_rates, std::vector<int> & starts, std::vector<int> & ends, int capacity, 
         std::map<int, double>& routing_probabilities); // for external lanes given schedule
    Lane(int id, int duration, double travel_time, std::vector<int> & starts, std::vector<int> & ends, int capacity, 
         std::map<int, double>& routing_probabilities); // for internal lanes given schedule
    
    Lane(int id, int duration, std::vector<double>& arrival_rates, std::vector<int> & starts, std::vector<int> & ends, int capacity, 
         std::map<int, double>& routing_probabilities, int queue_length, std::vector<Pair> & green_periods_so_far, 
         Pair & queue_state, Pair & lane_state, std::vector<int> & arrivals, int intention, int starting_second); // for external lanes given schedule + state
    Lane(int id, int duration, double travel_time, std::vector<int> & starts, std::vector<int> & ends, int capacity, 
         std::map<int, double>& routing_probabilities, int queue_length, std::vector<Pair> & green_periods_so_far, 
         Pair & queue_state, Pair & lane_state, std::vector<int> & arrivals, int intention, int starting_second); // for internal lanes given schedule + state
    
    void initialiseLastArrivalsInTheBeginning(Pair lane_state, std::vector<Lane> & lanes, int starting_second, bool is_empty, bool is_arrival);
    
    void updateQueues(int step);
    double getAverageQueue(int second);
    double getAverageSquareQueue(int second, int start_from) {return queues.getAverageSquareQueue(second, start_from);}
    double getAverageQueueInMeters(int second);
    std::vector<double> getNormalAverages(int duration);
    
    char getType() { return type; }
    double getFullTravelTime() { return travel_time; }
    int getShortTravelTime() { return short_travel_time; }
    void printTotalProbability(int second) {queues.checkAndPrintTotalProbability(second); queues.printQueues(second);}
    
    void updateLastArrivals(int step, const std::map<int, std::map<KeyDA, std::map<Pair, double>>>& tr_flows_zero,
                            const std::map<int, std::map<Pair, std::map<Pair, double>>>& tr_flows_plus, std::map<int, LastArrivals>& new_last_arrivals, 
                            const std::map<int, std::map<KeyDI, double>> & unblocking_probabilities);
    
    void updateLastArrivalsAfterDepartureToU(int step, int u, const std::map<KeyDA, std::map<Pair, double>>& tr_flows_zero,
                                             const std::map<Pair, std::map<Pair, double>>& tr_flows_plus, LastArrivals & new_last_arrivals, 
                                             const std::map<KeyDI, double> & unblocking_probabilities);

    
    std::map<KeyIA, double> getUnknownTransitionsProbabilities(int second);
    std::map<KeyDA, double> getZeroProbabilities(int second);
    std::map<char, double> getBTransitions(int second);
//     std::map<KeyDA, double> getJustEmptiedProbabilities(int second, std::map<char, std::map<char, double>>& ar_transitions);
    
//     std::map<char, double> getArTransitions(int second, char a_key);
    std::map<char, std::map<char, double>> getAllArTransitions(int second);
//     std::map<char, double> getArTransitionsFull(int second, char a_key);
    
//     std::map<Pair, Pair> getDefaultDepTransitions(int second, std::set<Pair>& d_keys);
//     bool isDepartureSecond(int second) {return departure_chain.isDepartureSecond(second); }
//     int getDeparuteNumber(int second) {return departure_chain.getDepartureNumber(second); }
    
    double getAverageArrivals(int from, int to);
    
    std::map<KeyIA, double> getBlockingProbabilities(int step);
    void generateTails(int step, double gamma, int u, std::vector<Tail> & tails, const std::map<KeyIA, double>& blocking_probabilities, 
                       Lane & downstream_lane, LastArrivals & new_last_arrivals,
                       std::map<KeyDA, std::map<Pair, double>>& tr_flows_zero, std::map<Pair, std::map<Pair, double>>& tr_flows_plus,
                       std::map<Pair, std::map<int, double>>& unblocking_d_keys_probabilities, std::map<KeyDI, double> & unblocking_probabilities,
                       std::vector<double> & unblocked_probabilities_of_tails, double travel_time);
    
    std::map<KeyIA, double> processTails(int step, std::vector<Tail> & tails, std::vector<double>& unblocked_probabilities_of_tails);
    
//     int getNextUnblock(Pair d_key);
    std::map<KeyIA, std::map<Pair, double>> getAlmostFullDKeys(int second);
    int getCapacity() { return capacity; } 
    
    void printArrivals(int second) { arrival_chain.printStates(second); }
    void printLastArrivals(int u) { last_arrivals[u].printLastArrivals(); }
    bool isGreen(int second) { return departure_chain.isGreen(second); }

    void processDepartures(int step, std::vector<Lane>& lanes);
    void processDeparturesOneLane(int step, std::vector<Lane>& lanes, int u, double gamma, std::vector<Tail> & tails, 
                                  std::vector<double> & unblocked_probabilities_of_tails, double travel_time);
    void preprocessLastArrivalsAndDepartureChain(int step, std::vector<Lane>& lanes, std::map<int, double> & travel_times);
    void updateAfterDepartures(int step,  std::map<int, std::map<KeyDA, std::map<Pair, double>>> & tr_flows_zero, 
                                 std::map<int, std::map<Pair, std::map<Pair, double>>> & tr_flows_plus, std::map<int, LastArrivals> & new_last_arrivals, 
                                 const std::map<int, std::map<KeyDI, double>> & unblocking_probabilities,
                                 const std::map<Pair, std::map<int, double>> & unblocking_d_keys_probabilities);
    void updateLastArrivalsAfterDepartureFromOtherLane(int step, int u, std::vector<Tail> & tails, std::map<KeyIA, std::vector<int>> & ordered_tails, 
                                                       std::vector<double> & unblocked_probabilities_of_tails);
    void updateLastArrivalsAfterAllDepartures(int step);
    void compareLastArrivals();
    
    void cleanUnprobableStates(int step);
    
    int getTotalNumberOfQueues();
    void printDistributionsToFile(int duration, const std::string & file_name);
    std::vector<double> getDistribution(int second) {return queues.getDistribution(second);}
    std::vector<std::vector<double>> getRightCumulativeDistributions(int start, int end) {return queues.getRightCumulativeDistributions(start, end);}
    
    double getAverageAndRemoveQueues(int second);
    double getAverageSquareAndRemoveQueues(int second, int start_from = 0);
    void printRoutingProbabilities();
};

// void multiplyLAByCoef(std::map<KeyIA, double> & la_probs, double coef);


#endif // LANE_H
