#include "lastarrivals.h"
#include "lane.h"


double LastArrivals::initialise(int step, std::map<KeyDA, double>& zero_probs, Lane& downsteam_lane){
    // clear old values
    double minus_first_probability;
    last_arrivals.clear();
    last_delayed_arrivals.clear();

    // get downstream states
    int travel_time = downsteam_lane.getShortTravelTime();
    int downstream_second = step + travel_time - 1;
    std::map<KeyIA, double> untransitioned = downsteam_lane.getUnknownTransitionsProbabilities(downstream_second);
    std::map<char, double> b_tr_probabilities = downsteam_lane.getBTransitions(downstream_second - 1);
    minus_first_probability = updatePBD(la_before_departure, step, downstream_second, untransitioned, b_tr_probabilities);
    
    
    // no need to add empty maps because pla and plda should be checked everywhere separately
    for (const auto & ia_iter: untransitioned){
        double total_zero = 0;
        KeyIA ia_key = ia_iter.first;
        double down_prob = ia_iter.second;
        for (const auto & da_iter: zero_probs){
            KeyDA da_key = da_iter.first;
            double prob = da_iter.second;
            total_zero += prob;
            last_arrivals[da_key][ia_key] = down_prob * prob;
        }
        if (1 - total_zero - minus_first_probability > 0){
            if (step > 0) { // it is done just before a green period
                la_before_departure[Pair(step + extra_green, 0)][ia_key] = down_prob * (1 - total_zero - minus_first_probability);
            } else { // in this case the first state maybe before (step, 0)
                auto da_iter = zero_probs.begin();
                int second = (da_iter->first).first.first;
                la_before_departure[Pair(second, 0)][ia_key] = down_prob * (1 - total_zero - minus_first_probability);
            }
        } else {
            if (total_zero + minus_first_probability > 1 + small_eps){
                std::cout << total_zero << " " << minus_first_probability << "\n";
                throw "Wrond total zero plus minus first probabilities!\n";
            }
        }
    }
    return minus_first_probability;
}


void LastArrivals::initialiseInTheBeginning(int step, Pair lane_state, Lane& downsteam_lane, bool queue_is_empty, bool is_arrival){
    KeyIA ia_key;
    double down_prob;
    int travel_time = downsteam_lane.getShortTravelTime();
    int downstream_second = step + travel_time - 1;
    std::map<KeyIA, double> untransitioned = downsteam_lane.getUnknownTransitionsProbabilities(downstream_second);
    
    // this is basically is not needed:
    last_arrivals.clear();
    last_delayed_arrivals.clear();
    la_before_departure.clear();
    
    // initialise either PLA or PBD
    if (queue_is_empty) {
        for (const auto & ia_iter: untransitioned){
            ia_key = ia_iter.first;
            down_prob = ia_iter.second;
            last_arrivals[KeyDA(lane_state, is_arrival ? 'a' : 'p')][ia_key] = down_prob;
        }
    } else {
        for (const auto & ia_iter: untransitioned){
            ia_key = ia_iter.first;
            down_prob = ia_iter.second;
            la_before_departure[lane_state][ia_key] = down_prob;
        }        
    }
}


std::map<KeyDA, double> getZeroProbs(LastArrivals & last_arrivals){
    std::map<KeyDA, double> probs_per_da_key;
    for (auto & da_iter: last_arrivals.last_arrivals){
        probs_per_da_key[da_iter.first] = sumMapValues(da_iter.second);
    }
    for (auto & da_iter: last_arrivals.last_delayed_arrivals){
        double sum = std::accumulate(da_iter.second.begin(), da_iter.second.end(), 0., [&](double value, const auto & pair) 
                                     { return value + sumMapValues(pair.second);});
        addValue(probs_per_da_key, da_iter.first, sum);
    }
    return probs_per_da_key;
}

std::map<Pair, double> getNonZeroProbs(const LastArrivals & last_arrivals){
    std::map<Pair, double> probs_per_d_key;
    for (auto & d_iter: last_arrivals.la_before_departure){
        probs_per_d_key[d_iter.first] = sumMapValues(d_iter.second);
    }
    return probs_per_d_key;
}

void printZeroAndNonZeroProbabilities(const std::map<KeyDA, double>& zero_probs, DepartureChain & departure_chain, LastArrivals & last_arrivals){
    last_arrivals.printLastArrivals();
    
    std::map<KeyDA, double> probs_per_da_key = getZeroProbs(last_arrivals), differences_zero;
    for (auto & da_iter: probs_per_da_key){
        differences_zero[da_iter.first] = probs_per_da_key[da_iter.first] - getValue(zero_probs, da_iter.first, 0.);
    }
    std::cout << "\tZero probs are\n" << probs_per_da_key << "\n but should be\n" << zero_probs << "\n";
    std::cout << "differences are:\n" <<  differences_zero << "\n";
    
    std::map<Pair, double> probs_per_d_key, should_be_per_d_key, differences;
    for (auto & d_iter: last_arrivals.la_before_departure){
        probs_per_d_key[d_iter.first] = sumMapValues(d_iter.second);
        should_be_per_d_key[d_iter.first] = departure_chain.getStateProbability(d_iter.first);
        differences[d_iter.first] = probs_per_d_key[d_iter.first] - should_be_per_d_key[d_iter.first];
    }
    std::cout << "\tNon-zero probs are\n" << probs_per_d_key << "\n but should be\n" << should_be_per_d_key << "\n";
    std::cout << "differences are:\n" <<  differences << "\n";
}

bool checkDifferentLastArrivals(std::map<int, LastArrivals> & last_arrivals){
    // should have the same total zero and non-zero probs
    // compare with the first
    if (last_arrivals.size() <= 1)
        return true; // everything is good (nothing to compare)
    int u_one = last_arrivals.begin()->first;
    std::map<KeyDA, double> probs_per_da_key, should_be_per_da_key = getZeroProbs(last_arrivals[u_one]);
    std::map<Pair, double> probs_per_d_key, should_be_per_d_key = getNonZeroProbs(last_arrivals[u_one]);
    
    for (auto & iter: last_arrivals){
        if (iter.first == u_one)
            continue;
        probs_per_da_key = getZeroProbs(iter.second);
        probs_per_d_key = getNonZeroProbs(iter.second);
        if (!mapsAlmostEqual(probs_per_da_key, should_be_per_da_key) || !mapsAlmostEqual(probs_per_d_key, should_be_per_d_key)){
            last_arrivals[u_one].printLastArrivals();
            iter.second.printLastArrivals(); 
            std::cout << "For " << u_one << " totals are (zero, non-zero) " << should_be_per_da_key << ";\n" << should_be_per_d_key << "\n";
            std::cout << "For " << iter.first << " totals are (zero, non-zero) " << probs_per_da_key << ";\n" << probs_per_d_key << "\n";
            std::cout << "Incorrect last arrivals for indexes " << u_one << " and " << iter.first << "!\n";
            return false;
        }
    }
    return true;
}

bool LastArrivals::checkTotalProbabilities(const std::map<KeyDA, double>& zero_probs, DepartureChain & departure_chain){
    double total_zero, every_zero = 0, plus_prob = 0, prob;
    for (auto & da_iter: last_arrivals){
        total_zero = 0;
        for (auto & ia_iter: da_iter.second) {
            prob = ia_iter.second;
            total_zero += prob;
            if (prob < 0){
                printLastArrivals();
                std::cout << prob << " " << da_iter << " " << ia_iter.first << "\n";
                std::cout << "Negative probability in PLA!\n";
                return false;
            }
        }
        if (isInMap(last_delayed_arrivals, da_iter.first)) {
            for (auto & iter: last_delayed_arrivals[da_iter.first]){
                for (auto & ia_iter: iter.second) {
                    prob = ia_iter.second;
                    total_zero += prob;
                    if (prob < 0){
                        printLastArrivals();
                        std::cout << prob << " " << da_iter << " " << ia_iter.first << "\n";
                        std::cout << "Negative probability in PLDA!\n";
                        return false;
                    }
                }
            }
        }
        if (((!isInMap(zero_probs, da_iter.first) && total_zero > small_eps) || !almostEqual(total_zero, getValue(zero_probs, da_iter.first, 0.), small_eps))){
            printLastArrivals();
            printZeroAndNonZeroProbabilities(zero_probs, departure_chain, *this);
            std::cout << "Incorrect total zero probability for last arrivals!\n";
            std::cout << "for " << da_iter.first << " " << total_zero << "\n";
            std::cout << "for " << da_iter.first << " it should be " << zero_probs.at(da_iter.first) << " (" << 1 - zero_probs.at(da_iter.first) 
                      << ") and it is " << total_zero << " (" << 1 - total_zero << ")"
                      << " (difference is " << zero_probs.at(da_iter.first) - total_zero << ")\n";
            std::cout << "Incorrect total zero probability for last arrivals!\n";
            return false;
        }
        every_zero += total_zero;
    }
    for (auto & da_iter: last_delayed_arrivals){
        if (isInMap(last_arrivals, da_iter.first))
            continue; // already checked
        total_zero = 0;
        for (auto & iter: last_delayed_arrivals[da_iter.first]){
            for (auto & ia_iter: iter.second) {
                prob = ia_iter.second;
                total_zero += prob;
                if (prob < 0){
                    printLastArrivals();
                    std::cout << prob << " " << da_iter << " " << ia_iter.first << "\n";
                    std::cout << "Negative probability in PLDA!\n";
                    return false;
                }
            }
        }
        if (!almostEqual(total_zero, getValue(zero_probs, da_iter.first, 0.), small_eps)){
            printLastArrivals();
            std::cout << "for " << da_iter.first << " it should be " << zero_probs.at(da_iter.first) << " and it is " << total_zero 
                      << " (difference is " << zero_probs.at(da_iter.first) - total_zero << ")\n";
            std::cout << "Incorrect total zero probability for last arrivals!\n";
            return false;
        }
        every_zero += total_zero;
    }
    for (auto & da_iter: zero_probs){
        if (!isInMap(last_arrivals, da_iter.first) && !isInMap(last_delayed_arrivals, da_iter.first)){
            if (da_iter.second > small_eps && da_iter.first.first.second >= 0){
                printLastArrivals();
                std::cout << da_iter.first << " " << da_iter.second << "\n";
                std::cout << "No such states in last arrivals!\n";
                return false;
            }
        }
    }
    for (auto & d_iter: la_before_departure){
        double d_prob = 0;
        for (auto & ia_iter: d_iter.second) {
            prob = ia_iter.second;
            plus_prob += prob;
            d_prob += prob;
            if (prob < 0){
                printLastArrivals();
                std::cout << prob << " " << d_iter.first << " " << ia_iter.first << "\n";
                std::cout << "Negative probability in PBD!\n";
                return false;
            }
        }
        double d_prob_should_be = departure_chain.getStateProbability(d_iter.first);
        if (d_iter.first.second >= 0 && !almostEqual(d_prob, d_prob_should_be, small_eps)){ // do not compare for red states
            printLastArrivals();
            printZeroAndNonZeroProbabilities(zero_probs, departure_chain, *this);
            std::cout << "probability of " << d_iter.first << " should be " << d_prob_should_be  << " (" << 1 - d_prob_should_be << ")" 
                      << ", but it is " << d_prob << " (" <<  1 - d_prob << ")" 
                      << " (the difference is " << d_prob_should_be - d_prob << ")\n";
            std::cout << "Incorrect total probability of last arrivals for d_key!\n";
            return false;
        }
    }
    if (!almostEqual(every_zero + plus_prob, 1, small_eps)){
        printLastArrivals();
        std::cout << every_zero << " " << plus_prob << " " << 1 - (every_zero + plus_prob) << "\n";     
        std::cout << 1 - getTotalProbability() << "\n";
        std::cout << "Incorrect total probability of last arrivals!\n";
        return false;
    }
    return true;
}


bool LastArrivals::checkTotalProbabilities(DepartureChain& departure_chain){
    for (auto & d_iter: la_before_departure){
        double d_prob = 0;
        for (auto & ia_iter: d_iter.second) {
            double prob = ia_iter.second;
            d_prob += prob;
            if (prob < 0){
                printLastArrivals();
                std::cout << prob << " " << d_iter.first << " " << ia_iter.first << "\n";
                std::cout << "Negative probability in PBD!\n";
                return false;
            }
        }
        double d_prob_should_be = departure_chain.getStateProbability(d_iter.first);
        if (d_iter.first.second >= 0 && !almostEqual(d_prob, d_prob_should_be, small_eps)){ // do not compare for red states
            printLastArrivals();
            std::cout << "probability of " << d_iter.first << " should be " << d_prob_should_be  << " (" << 1 - d_prob_should_be << ")" 
                      << ", but it is " << d_prob << " (" <<  1 - d_prob << ")" 
                      << " (the difference is " << d_prob_should_be - d_prob << ")\n";
            std::cout << "Incorrect total probability of last arrivals for d_key!\n";
            return false;
        }
    }
    return true;
}


double LastArrivals::getTotalProbability(){
    auto sum_last_arrivals = [](auto & map) 
                {return std::accumulate(map.begin(), map.end(), 0., [&](double value, const auto & par){return value + sumMapValues(par.second);});};
    double total = sum_last_arrivals(last_arrivals) + sum_last_arrivals(la_before_departure);
    for (auto & da_iter: last_delayed_arrivals){
        total += sum_last_arrivals(da_iter.second);
    }
    return total;
}


void updateLastDelayedArrivals(std::map<int, MapIAProb>& last_delayed_arrivals_da){
    for (auto plda_iter = last_delayed_arrivals_da.begin(); plda_iter != last_delayed_arrivals_da.end(); /*no increment*/){
        int i_not_earlier = plda_iter->first;
        for (auto ia_iter = plda_iter->second.begin(); ia_iter != plda_iter->second.end(); /*no increment*/){
            int i_to = ia_iter->first.first;
            char a_key = ia_iter->first.second;
            if (i_to >= i_not_earlier){
                addValue(last_delayed_arrivals_da[(a_key == 'a') ? i_to + 2: i_to + 1], ia_iter->first, ia_iter->second);
                ia_iter = plda_iter->second.erase(ia_iter);
            } else {
                ++ia_iter;
            }
        }
        if (plda_iter->second.size() == 0)
            plda_iter = last_delayed_arrivals_da.erase(plda_iter);
        else
            ++plda_iter;
    }
}

void LastArrivals::updateNewSecond(int step, DepartureChain & departure_chain, const std::map<char, std::map<char, double>>&  ar_transitions,  
                                   const std::map<KeyDA, double> & just_emptied_probs, Lane& downsteam_lane, double real_travel_time){
    int travel_time = downsteam_lane.getShortTravelTime();
    int last_downstream_second = step + travel_time - 1;
    travel_time = real_travel_time;
    std::map<char, double> b_tr_probabilities = downsteam_lane.getBTransitions(last_downstream_second - 1);
    // update old values
    for (auto & da_iter: last_arrivals){  // should be the same as     for (auto & da_iter: last_delayed_arrivals){
        updateToNewSecond(last_downstream_second, last_arrivals[da_iter.first], b_tr_probabilities);
    }
    for (auto & da_iter: last_delayed_arrivals){
        for (auto & iter: last_delayed_arrivals[da_iter.first]){
            updateToNewSecond(last_downstream_second, last_delayed_arrivals[da_iter.first][iter.first], b_tr_probabilities);
        }
        updateLastDelayedArrivals(last_delayed_arrivals[da_iter.first]);
    }
    for (auto & d_iter: la_before_departure){
        updateToNewSecond(last_downstream_second, la_before_departure[d_iter.first], b_tr_probabilities);
    }
    
    
    // mix according to transition probabilities
    std::map<KeyDA, MapIAProb> new_last_arrivals; 
    std::map<KeyDA, std::map<int, MapIAProb>> new_last_delayed_arrivals; 
    KeyDA da_key, new_da_key;
    KeyIA ia_key;
    Pair d_key, new_d_key;
    char a_key, new_a_key;
    double prob, d_state_prob;
    for (auto & da_iter: last_arrivals){
        da_key = da_iter.first;
        d_key = da_key.first;
        new_d_key = departure_chain.getDefaultZeroTransition(step - 1, da_key.first);
        while (departure_chain.getLastSecond(new_d_key) < step)
            new_d_key = departure_chain.getDefaultZeroTransition(step - 1, new_d_key);
//         // NOTE: here the transition is different than the default transition!!!
        a_key = da_key.second;
        for (auto & ia_iter: last_arrivals[da_key]){
            if (ia_iter.second < almost_zero || (ia_iter.second < small_eps && !isInMap(ar_transitions, a_key)))
                continue; // do not consider low probability 
            ia_key = ia_iter.first;
            for (auto & a_iter: ar_transitions.at(a_key)){
                new_a_key = a_iter.first;
                prob = a_iter.second;
                new_da_key = KeyDA(new_d_key, new_a_key);
                addValue(new_last_arrivals, new_da_key, ia_key, ia_iter.second * prob);
            }
        }
    }
    for (auto & da_iter: last_delayed_arrivals){
        da_key = da_iter.first;
        d_key = da_key.first;
        new_d_key = departure_chain.getDefaultZeroTransition(step - 1, da_key.first);
        while (departure_chain.getLastSecond(new_d_key) < step)
            new_d_key = departure_chain.getDefaultZeroTransition(step - 1, new_d_key);
//         // NOTE: here the transition is different than the default transition!!!
        a_key = da_key.second;
        for (const auto & iter: da_iter.second){
            for (const auto & ia_iter: iter.second){
                if (ia_iter.second < almost_zero || (ia_iter.second < small_eps && (!isInMap(ar_transitions, a_key))))
                    continue;
                ia_key = ia_iter.first;
                for (auto & a_iter: ar_transitions.at(a_key)){
                    new_a_key = a_iter.first;
                    prob = a_iter.second;
                    new_da_key = KeyDA(new_d_key, new_a_key);
                    if (isInMap(new_last_delayed_arrivals, new_da_key))
                        addValue(new_last_delayed_arrivals[new_da_key], iter.first, ia_key, ia_iter.second * prob);
                    else 
                        new_last_delayed_arrivals[new_da_key][iter.first][ia_key] = ia_iter.second * prob;
                }
            }
        }
    }
    
    last_arrivals = new_last_arrivals;
    last_delayed_arrivals = new_last_delayed_arrivals;
    
    
    std::map<Pair, double> emptied_d_key;
    std::map<Pair, double> d_key_probs;
    int dep_index, new_i_int;
    double new_i;
    for (auto & da_iter: just_emptied_probs){
        da_key = da_iter.first;
        d_key = da_key.first;
        if (da_iter.second > almost_zero)
            addValue(d_key_probs, d_key, da_iter.second);
    }
    
    for (auto & d_iter: d_key_probs){
        d_iter.second += departure_chain.getStateProbability(d_iter.first);
    } // now d_key_probs[d_key] = sum_{ia_key} la_before_departure[d_key][ia_key]
    
    
    for (const auto & da_iter: just_emptied_probs){
        if (da_iter.second < almost_zero)
            continue; // no need to process
        da_key = da_iter.first;
        d_key = da_key.first;
        if (departure_chain.isSplitLane()){
            da_key.first = Pair(0, 0); // this PLDA should be updated
        }
        dep_index = d_key.second - 1;
        new_i = departure_chain.getLastSecond(Pair(d_key.first, dep_index)) + travel_time + getArrivalMoment(dep_index) - getDepartureMoment(dep_index) + 2;
        new_i_int = int(new_i);
        d_state_prob = d_key_probs[d_key];
        if (d_state_prob > almost_zero) { // departures are possible
            prob = da_iter.second / d_state_prob;
            if (!isInMap(la_before_departure, d_key)){
                if (da_iter.second > small_eps){
                    printLastArrivals();
                    std::cout << da_key << " <- processing da_key in just emptied\n";
                    std::cout << da_iter.second << "\n";
                    throw "More emptied than could\n";
                } else {
                    continue; // skip low probability mistake
                }
            }
            addValue(emptied_d_key, d_key, prob);
            for (const auto & ia_iter: la_before_departure.at(d_key)){
                ia_key = ia_iter.first;
                if (isInMap(last_delayed_arrivals, da_key)){
                    addValue(last_delayed_arrivals[da_key], new_i_int, ia_key, la_before_departure[d_key][ia_key] * prob);
                } else {
                    last_delayed_arrivals[da_key][new_i_int][ia_key] = la_before_departure[d_key][ia_key] * prob;
                }
            }
        }
    }
    for (const auto & d_iter: emptied_d_key){
        d_key = d_iter.first;
        d_state_prob = d_key_probs[d_key];
        if (emptied_d_key[d_key] * d_state_prob > small_eps + d_state_prob){
            // more emptied than was possible
            std::cout << emptied_d_key << " <- emptied\n";
            std::cout << d_key << "\n";
            std::cout << emptied_d_key[d_key] << " " << d_state_prob << "\n";
            throw "More emptied than was possible when updating last arrivals probabilities.\n";
        }
        emptied_d_key[d_key] = std::min(emptied_d_key[d_key], 1.);
        for (const auto & ia_iter: la_before_departure[d_key]){
            ia_key = ia_iter.first;
            la_before_departure[d_key][ia_key] *= (1 - emptied_d_key[d_key]);
        }
    }
}


double pushToNewSecond(int new_second, MapIAProb& my_map, MapIAProb & untransitioned, const std::map<char, double> & b_tr_probabilities){
    // it is similar to updateToNewSecond but it is pushed from any last second and without b_probs
    
    double total = 0.;
    char a_key;
    for (auto iter = my_map.begin(); iter != my_map.end(); /* no increment */){
        KeyIA ia_key = iter->first;
        int i = ia_key.first;
        double prob = iter->second;
        if (i < new_second){ 
            if (ia_key.second == 'b'){
                for (auto & b_to_iter: b_tr_probabilities){
                    a_key = b_to_iter.first;
                    addValue(my_map, KeyIA(new_second, a_key), prob * b_to_iter.second);
                }
            } else {
                a_key = default_ar_transitions.at(ia_key.second);
                addValue(my_map, KeyIA(new_second, a_key), prob);
            }
            iter = my_map.erase(iter); // points to the next element
        } else { // everything is fine just continue
            ++iter;
        }
    }
    for (auto & iter: my_map){
        double prob = iter.second;
        double value = untransitioned[iter.first];
        if (prob > value){
            prob = value;
            iter.second = value; 
            // TODO? check why this may be the case (with big difference)
        }
        total += prob;
        addValue(untransitioned, iter.first, -prob); // thus it is already used
    }
    return total;
}

void updateToNewSecond(int second, MapIAProb& my_map, const std::map<char, double> & b_probs){
    for (auto iter = my_map.begin(); iter != my_map.end(); /* no increment */){
        KeyIA ia_key = iter->first;
        int i = ia_key.first;
        double prob = iter->second;
        if (i < second){  // should be then second - 1
            char a_key = ia_key.second;
            if (a_key == 'b'){
                for (const auto & b_tr: b_probs){
                    addValue(my_map, KeyIA(i + 1, b_tr.first), prob * b_tr.second);
                }
            } else { // just add to the default transition of a_key
                a_key = default_ar_transitions.at(a_key);
                addValue(my_map, KeyIA(i + 1, a_key), prob);
            }
            iter = my_map.erase(iter); // points to the next element
        } else { // everything is fine just continue
            ++iter;
        }
    }
}

bool updateIAMapFromTails(std::map<KeyIA, std::vector<int>> & ordered_tails, const std::vector<Tail>& tails, MapIAProb & last_arrivals, const MapIAProb & totals, 
                          const std::vector<double> & unblocked_probabilities_of_tails){
    MapIAProb new_keys;
    for (auto & ia_iter: last_arrivals){
        KeyIA ia_key = ia_iter.first;
        double total = totals.at(ia_key);
        double changed = 0;
        double prob = ia_iter.second;
        if (total > almost_zero && isInMap(ordered_tails, ia_key)){
            for (unsigned int i = 0; i < ordered_tails[ia_key].size(); ++i){
                const Tail & tail = tails.at(ordered_tails[ia_key][i]);
                double new_prob = 0;
                if (tail.a_key_to != 'b') {
                    new_prob = tail.probability * prob / total;
                    addValue(new_keys, KeyIA(tail.second_to, tail.a_key_to), new_prob);
                } else {
                    double unblocked = unblocked_probabilities_of_tails.at(ordered_tails[ia_key][i]);
                    if (unblocked < tail.probability){
                        new_prob = (tail.probability - unblocked) * prob / total;
                        addValue(new_keys, KeyIA(tail.second_to, tail.a_key_to), new_prob);
                    }
                    if (unblocked > 0) {
                        new_prob = unblocked * prob / total;
                        addValue(new_keys, KeyIA(tail.second_to, 'p'), new_prob);
                    }
                }
                changed += new_prob;
            }
        }
        if (prob > changed){
            addValue(new_keys, ia_key, prob - changed);
        } else if (changed > prob + small_eps) {
            std::cout << totals << ";\n" << last_arrivals << "\n";
            std::cout << ia_key << " " << total << " " << prob << " " << changed << " " << prob - changed << "\n"; 
            throw "updateIAMapFromTails: more changed than the last arrival probability!\n";
        }
    }
    last_arrivals = new_keys;
    return (new_keys.size() == 0);
}

template <class key_type>
bool updateIAMaps(std::map<key_type, MapIAProb> & last_arrivals, std::map<KeyIA, std::vector<int>> & ordered_tails, std::vector<Tail>& tails, 
                                   std::vector<double> & unblocked_probabilities_of_tails, MapIAProb & totals){
    for (auto key_iter = last_arrivals.begin(); key_iter != last_arrivals.end(); /*no increment*/){
        bool erase = updateIAMapFromTails(ordered_tails, tails, key_iter->second, totals, unblocked_probabilities_of_tails);
        if (erase) {
            key_iter = last_arrivals.erase(key_iter);
        } else {
            ++key_iter;
        }
    }
    return (last_arrivals.size() == 0);
}

void LastArrivals::updateFromTails(std::map<KeyIA, std::vector<int>> & ordered_tails, std::vector<Tail>& tails, 
                                   std::vector<double> & unblocked_probabilities_of_tails){
    MapIAProb totals;
    
    // compute totals
    for (auto & da_iter: last_arrivals){
        addMap(totals, da_iter.second);
    }
    for (auto & da_iter: last_delayed_arrivals){
        for (auto & i_iter: da_iter.second){
            addMap(totals, i_iter.second);
        }
    }
    for (auto & d_iter: la_before_departure){
        addMap(totals, d_iter.second);
    }
    
    // compute new maps
    updateIAMaps(last_arrivals, ordered_tails, tails, unblocked_probabilities_of_tails, totals);
    for (auto da_iter = last_delayed_arrivals.begin(); da_iter != last_delayed_arrivals.end(); /*no_increment*/){
        bool erase = updateIAMaps(da_iter->second, ordered_tails, tails, unblocked_probabilities_of_tails, totals);
        if (erase) {
            da_iter = last_delayed_arrivals.erase(da_iter);
        } else {
            ++da_iter;
        }
    }
    updateIAMaps(la_before_departure, ordered_tails, tails, unblocked_probabilities_of_tails, totals);
}


double updatePBD(std::map<Pair, MapIAProb> & la_before_departure, int second, int downstream_second, std::map<KeyIA, double> & untransitioned,
                 const std::map<char, double> & b_tr_probabilities){
    Pair d_key;
    double total_blocked = 0;
    for (auto d_iter = la_before_departure.begin(); d_iter != la_before_departure.end(); /* no increment */){
        d_key = d_iter->first;
        if (d_key.first <= second +  extra_green) {
            d_iter = la_before_departure.erase(d_iter);
        } else {
            total_blocked += pushToNewSecond(downstream_second, d_iter->second, untransitioned, b_tr_probabilities);
            ++d_iter;
        }
    }
    if (1 - total_blocked > almost_zero){
        // this is NOT wrong: the blocked probabilities are already taken from the untransitioned
        // but later I multiply by 1 - total_blocked
        double coef = 1 - total_blocked;
        for (auto & ai_iter: untransitioned){
            ai_iter.second /= coef;
        }
    } else {
        // everything is blocked
        untransitioned.clear();
    }
    return total_blocked;
}


void LastArrivals::printLastArrivals(){
    char a_key, a_key_bar;
    Pair d_key;
    int i, i_del;
    double total = 0;
    std::cout << "\tPLA\n";
    for (auto & da_iter: last_arrivals){
        d_key = da_iter.first.first;
        a_key = da_iter.first.second;
        std::cout << d_key << ", " << a_key << ":\n";
        for (auto & ia_iter: da_iter.second){
            a_key_bar = ia_iter.first.second;
            i = ia_iter.first.first;
            std::cout << d_key << ", " << a_key << ", " << i << ", " << a_key_bar << " -> " << ia_iter.second << std::endl;
            total += ia_iter.second;
        }
    }
    std::cout << "\tPLDA\n";
    for (auto & da_iter: last_delayed_arrivals){
        d_key = da_iter.first.first;
        a_key = da_iter.first.second;
        std::cout << d_key << ", " << a_key << ":\n";
        for (auto & iter: da_iter.second) { 
            i_del = iter.first;
            for (auto & ia_iter: iter.second){
                a_key_bar = ia_iter.first.second;
                i = ia_iter.first.first;
                std::cout << d_key << ", " << a_key << ", " << i << ", " << a_key_bar << ", " << i_del <<" -> " << ia_iter.second << std::endl;
                total += ia_iter.second;
            }
        }
    }
    std::cout << "\tPBD\n";
    for (auto & d_iter: la_before_departure){
        d_key = d_iter.first;
        std::cout << d_key << ":\n";
        for (auto & ia_iter: d_iter.second){
            a_key_bar = ia_iter.first.second;
            i = ia_iter.first.first;
            std::cout << d_key << ", " << i << ", " << a_key_bar << " -> " << ia_iter.second << std::endl;
            total += ia_iter.second;
        }
    }
    std::cout << "together: " << total << " (" << 1 - total << ")\n";
}

void LastArrivals::removeStates(const std::vector<Pair> & d_keys_to_remove){
    for (auto d_key: d_keys_to_remove) {
        for (auto da_iter = last_arrivals.begin(); da_iter != last_arrivals.end(); /*no increment*/){
            if (da_iter->first.first != d_key) {
                ++da_iter;
            } else {
                da_iter = last_arrivals.erase(da_iter);
            }
        }
        for (auto da_iter = last_delayed_arrivals.begin(); da_iter != last_delayed_arrivals.end(); /*no increment*/){
            if (da_iter->first.first != d_key) {
                ++da_iter;
            } else {
                da_iter = last_delayed_arrivals.erase(da_iter);
            }
        }
        for (auto d_iter = la_before_departure.begin(); d_iter != la_before_departure.end(); /*no increment*/){
            if (d_iter->first != d_key) {
                ++d_iter;
            } else {
                d_iter = la_before_departure.erase(d_iter);
            }
        }
    }
}
