#include <fstream>
#include <iomanip>
#include "network.h"

const bool TO_DELETE = false;

void read_state(std::ifstream & state_file, int id, int & queue_length, std::vector<Pair> & green_periods_so_far, 
                Pair & queue_state, Pair & lane_state, std::vector<int> & arrivals, int & intention){
    int new_id, number_of_green_periods, number_of_arrivals, s, k;
    state_file >> new_id >> queue_length >> number_of_green_periods;
    if (new_id != id){
        std::cout << id << " " << new_id << "\n";
        throw "Incorrect state or lanes file\n";
    }
    green_periods_so_far.reserve(number_of_green_periods);
    for (int i = 0; i < number_of_green_periods; ++ i){
        state_file >> s >> k;
        green_periods_so_far.emplace_back(s, k);
    }
    state_file >> s >> k;
    queue_state = Pair(s, k);
    state_file >> s >> k;
    lane_state = Pair(s, k);
    state_file >> number_of_arrivals;
    arrivals.reserve(number_of_arrivals);
    for (int i = 0; i < number_of_arrivals; ++ i){
        state_file >> s;
        arrivals.emplace_back(s);
    }
    state_file >> intention;
}

void initialiseExternalLane(int id, int duration, double lamb, double factor, int cycle, int offset, int red, int green, int length, 
                            std::map<int, double> routing_probabilities, std::vector<Lane>& lanes){
    std::vector<double> arrival_rates;
    int shift = std::ceil(length / free_flow_speed);     arrival_rates.resize(shift, 0); // no arrivals yet
    arrival_rates.resize(duration / 2 + shift, lamb * factor);
    arrival_rates.resize(duration, lamb * (2 - factor));
    lanes.emplace_back(id, duration, arrival_rates, cycle, offset, red, green, length / bumper_to_bumper, routing_probabilities);
}

void initialiseExternalLane(int id, int duration, double lamb, double factor, std::vector<int> & green_starts, std::vector<int> & green_ends, 
                            double length, int capacity, std::map<int, double> routing_probabilities, std::vector<Lane>& lanes){
    std::vector<double> arrival_rates;
    int shift = std::ceil(length / free_flow_speed);     arrival_rates.resize(shift, 0); // no arrivals yet
    arrival_rates.resize(duration / 2 + shift, lamb * factor);
    arrival_rates.resize(duration, lamb * (2 - factor));
    lanes.emplace_back(id, duration, arrival_rates, green_starts, green_ends, capacity, routing_probabilities);
}

void initialiseExternalLane(int id, int duration, double lamb, double factor, std::vector<int> & green_starts, std::vector<int> & green_ends, 
                            double length, int capacity, std::map<int, double> routing_probabilities, std::vector<Lane>& lanes, 
                            std::ifstream & state_file, int starting_second, Pair & lane_state, bool & is_lane_empty, bool & is_lane_arrival){
    std::vector<double> arrival_rates;
    int shift = std::ceil(length / free_flow_speed);     int total_duration = starting_second + duration;
    arrival_rates.resize(shift, 0); // no arrivals yet
    arrival_rates.resize(total_duration / 2 + shift, lamb * factor);
    arrival_rates.resize(total_duration, lamb * (2 - factor));
    int queue_length, intention;
    Pair queue_state;
    std::vector<Pair> green_periods_so_far;
    std::vector<int> arrivals;
    read_state(state_file, id, queue_length, green_periods_so_far, queue_state, lane_state, arrivals, intention);
    capacity = std::max(capacity, queue_length + int(arrivals.size()));
    is_lane_empty = (queue_length == 0);
    is_lane_arrival = (arrivals.size() > 0 && arrivals[0] == starting_second);
    lanes.emplace_back(id, duration, arrival_rates, green_starts, green_ends, capacity, routing_probabilities, 
                       queue_length, green_periods_so_far, queue_state, lane_state, arrivals, intention, starting_second);
}

void initialiseInternalLane(int id, int duration, int travel_time, int cycle, int offset, int red, int green, int length, 
                            std::map<int, double> routing_probabilities, std::vector<Lane>& lanes){
    lanes.emplace_back(id, duration, travel_time, cycle, offset, red, green, length / bumper_to_bumper - 2, routing_probabilities); // TODO: check what exactly minus in capacity
}


void initialiseInternalLane(int id, int duration, int travel_time, std::vector<int> & green_starts, std::vector<int> & green_ends, int capacity, 
                            std::map<int, double> routing_probabilities, std::vector<Lane>& lanes){    
    lanes.emplace_back(id, duration, travel_time, green_starts, green_ends, capacity, routing_probabilities); 
}

void initialiseInternalLane(int id, int duration, int travel_time, std::vector<int> & green_starts, std::vector<int> & green_ends, int capacity, 
                            std::map<int, double> routing_probabilities, std::vector<Lane>& lanes, std::ifstream & state_file, int starting_second, 
                            Pair & lane_state, bool & is_lane_empty, bool & is_lane_arrival){
    int queue_length, intention;
    Pair queue_state;
    std::vector<Pair> green_periods_so_far;
    std::vector<int> arrivals;
    read_state(state_file, id, queue_length, green_periods_so_far, queue_state, lane_state, arrivals, intention);
    capacity = std::max(capacity, queue_length + int(arrivals.size()));
    is_lane_empty = (queue_length == 0);
    is_lane_arrival = (arrivals.size() > 0 && arrivals[0] == starting_second);
    lanes.emplace_back(id, duration, travel_time, green_starts, green_ends, capacity, routing_probabilities, 
                       queue_length, green_periods_so_far, queue_state, lane_state, arrivals, intention, starting_second); 
}


Network::Network(const std::string & lane_file_name, const std::string & arrival_rates_file_name, const std::string & routing_file_name, 
                 const std::string & green_times_file_name/*, std::vector<std::vector<int>> green_starts, std::vector<std::vector<int>> green_ends*/){
    std::ifstream lane_file(lane_file_name);
    std::ifstream rates_file(arrival_rates_file_name);
    std::ifstream routing_file(routing_file_name);
    std::ifstream green_file(green_times_file_name);
    int index, size, array_size;
    char lane_type;
    int duration = DURATION;
    int u_up, u_down, u;
    int moment;
    double lamb, factor = 1.;
    std::map<int, std::map<int, double>> routing_probabilities;
    std::map<int, double> arrival_rates;
    double length, gamma, travel_time;
    std::vector<std::vector<int>> green_starts, green_ends;
    
    std::cout << std::setprecision(16);
    
    starting_second = 0;
    
    // read arrival rates
    rates_file >> size; // dummy
    while (rates_file >> index >> lamb){
        arrival_rates[index] = lamb;
    }
    // read routing_file
    routing_file >> size;
    routing_probabilities_vector.reserve(size);
    while (routing_file >> u_up >> u_down >> gamma >> travel_time){
        routing_probabilities_vector.emplace_back(u_up, u_down, gamma);
        routing_probabilities[u_up][u_down] = gamma;
        travel_times[u_up][u_down] = travel_time;
    }
    // find number of lanes
    lane_file >> size;
    lanes.reserve(size);
    
    // read greens
    green_starts.resize(size);
    green_ends.resize(size);
    green_file >> size; 
    for (int i = 0; i < size; ++i){
        // read one line
        green_file >> u >> array_size;
        for (int j = 0; j < array_size; ++j){
            green_file >> moment;
            if (j % 2 == 0)
                green_starts[u].push_back(moment);
            else
                green_ends[u].push_back(moment);
        }
    }
    
    // read lanes
    while (lane_file >> index >> lane_type >> length){
        int capacity = int(length / bumper_to_bumper);
        if (lane_type == 'i'){
            initialiseInternalLane(index, duration, length / free_flow_speed, green_starts[index], green_ends[index], 
                                   capacity, routing_probabilities[index], lanes);
        } else if (lane_type == 'e') {
            lamb = arrival_rates[index];
            capacity = int(length - 44 / bumper_to_bumper);  // 44 meters for adding new quick arrivals
            initialiseExternalLane(index, duration, lamb, factor, green_starts[index], green_ends[index], length, 
                                   capacity, routing_probabilities[index], lanes);
        } else {
            std::cout << lane_type << "\n";
            throw "Incorrect lane type\n";
        }
    }
    
}


Network::Network(const std::string & lane_file_name, const std::string & arrival_rates_file_name, const std::string & routing_file_name, 
                 const std::string & green_times_file_name, const std::string & state_file_name
                 ){
    std::ifstream lane_file(lane_file_name);
    std::ifstream rates_file(arrival_rates_file_name);
    std::ifstream routing_file(routing_file_name);
    std::ifstream green_file(green_times_file_name);
    std::ifstream state_file(state_file_name);
    int index, size, array_size;
    char lane_type;
    int duration = DURATION;
    int u_up, u_down, u;
    int moment;
    double lamb, factor = 1.;
    std::map<int, std::map<int, double>> routing_probabilities;
    std::map<int, double> arrival_rates;
    double length, gamma, travel_time;
    std::vector<std::vector<int>> green_starts, green_ends;
    
    std::cout << std::setprecision(16);
    
    // read arrival rates
    rates_file >> size; // dummy
    while (rates_file >> index >> lamb){
        arrival_rates[index] = lamb;
    }
    // read routing_file
    routing_file >> size;
    routing_probabilities_vector.reserve(size);
    while (routing_file >> u_up >> u_down >> gamma >> travel_time){
        routing_probabilities_vector.emplace_back(u_up, u_down, gamma);
        routing_probabilities[u_up][u_down] = gamma;
        travel_times[u_up][u_down] = travel_time;
    }
    // find number of lanes
    lane_file >> size;
    lanes.reserve(size);
    
    // read greens
    green_starts.resize(size);
    green_ends.resize(size);
    green_file >> size; 
    for (int i = 0; i < size; ++i){
        // read one line
        green_file >> u >> array_size;
        for (int j = 0; j < array_size; ++j){
            green_file >> moment;
            if (j % 2 == 0)
                green_starts[u].push_back(moment);
            else
                green_ends[u].push_back(moment);
        }
    }
    
    // start reading state_file
    state_file >> starting_second >> size;
    
    // read lanes
    std::vector<Pair> lane_states(size, Pair(-1, -1));
    std::vector<bool> is_empty(size, true);
    std::vector<bool> is_arrival(size, false);
    Pair lane_state;
    bool is_lane_arrival, is_lane_empty;
    while (lane_file >> index >> lane_type >> length){
        int capacity = int(length / bumper_to_bumper);
        if (lane_type == 'i'){
            initialiseInternalLane(index, duration, length / free_flow_speed, green_starts[index], green_ends[index], 
                                   capacity, routing_probabilities[index], lanes, state_file, starting_second, lane_state, is_lane_empty, is_lane_arrival);
        } else if (lane_type == 'e') {
            lamb = arrival_rates[index];
            capacity = int(length - 44 / bumper_to_bumper);  // 44 meters for adding new quick arrivals
            initialiseExternalLane(index, duration, lamb, factor, green_starts[index], green_ends[index], length, 
                                   capacity, routing_probabilities[index], lanes, state_file, starting_second, lane_state, is_lane_empty, is_lane_arrival);
        } else {
            std::cout << lane_type << "\n";
            throw "Incorrect lane type\n";
        }
        lane_states[index] = lane_state;
        is_empty[index] = is_lane_empty;
        is_arrival[index] = is_lane_arrival;
    }
    
    for (int i = 0; i < size; ++i) {
        lanes[i].initialiseLastArrivalsInTheBeginning(lane_states[i], lanes, starting_second, is_empty[i], is_arrival[i]);
    }
}


Network::Network(int duration, int cycle, int green, int small_green, int yellow, int offset, 
                 double lamb, double minor_lamb, double factor, double pr_north, double pr_east){
    int length = 500, length2 = 300;
    int red = green + yellow;
    starting_second = 0;
    
    std::map<int, double> routing_probabilities;
    lanes.reserve(4);
    routing_probabilities[2] = 1 - pr_north;
    routing_probabilities[3] = pr_north;
    routing_probabilities_vector.emplace_back(0, 2, 1 - pr_north);
    initialiseExternalLane(0, duration, lamb, factor, cycle, 0, 0, green, length, routing_probabilities, lanes);
    
    routing_probabilities[2] = pr_east;
    routing_probabilities[3] = 1 - pr_east;
    routing_probabilities_vector.emplace_back(1, 3, 1 - pr_east);
    routing_probabilities_vector.emplace_back(0, 3, pr_north);
    routing_probabilities_vector.emplace_back(1, 2, pr_east);
    initialiseExternalLane(1, duration, minor_lamb, factor, cycle, 0, red, small_green, length, routing_probabilities, lanes);
    
    routing_probabilities.clear();
    initialiseInternalLane(2, duration, std::ceil(length2 / free_flow_speed), cycle, offset, 0, green, length2, routing_probabilities, lanes);
    
    initialiseInternalLane(3, duration, std::ceil(length2 / free_flow_speed), cycle, offset, 0, green, length2, routing_probabilities, lanes);
}


void computeOrderedTails(std::vector<Tail> & tails, std::map<KeyIA, std::vector<int>> & ordered_tails){
    for (unsigned int i = 0; i < tails.size(); ++i){
        if (tails[i].probability > almost_zero){
            ordered_tails[KeyIA(tails[i].second_from, tails[i].a_key_from)].push_back(i);
        }
    }
}


void Network::updateDepartures(int step){
    int lanes_number = lanes.size();
    for (int u = 0; u < lanes_number; ++u){
        lanes[u].preprocessLastArrivalsAndDepartureChain(step, lanes, travel_times[u]);
    }
    for (auto& route_iter: routing_probabilities_vector){
        int u = route_iter.u_from;
        int u_down = route_iter.u_to;
        double gamma = route_iter.probability;
        
        std::vector<Tail> tails;
        std::vector<double> unblocked_probabilities_of_tails;
        std::map<KeyIA, std::vector<int>> ordered_tails; // ordered by KeyIA
        
        lanes[u].processDeparturesOneLane(step, lanes, u_down, gamma, tails, unblocked_probabilities_of_tails, travel_times[u][u_down]);
        computeOrderedTails(tails, ordered_tails);
        for (int u_other = 0; u_other < lanes_number; ++ u_other){ // this is for the case of multiple lanes with the same downstream lane
            if (u_other == u)
                continue;
            lanes[u_other].updateLastArrivalsAfterDepartureFromOtherLane(step, u_down, tails, ordered_tails, unblocked_probabilities_of_tails);
        }
    }
    for (int u = 0; u < lanes_number; ++u){
        lanes[u].updateLastArrivalsAfterAllDepartures(step);
    }
}


int getStartFrom(int u, std::map<int, std::map<int, double>> & travel_times){
    if (isInMap(travel_times, u) && travel_times[u].size() > 1) { // this is a split lane
        double time = travel_times[u].begin()->second;
        return int(time * free_flow_speed / bumper_to_bumper);
    }
    return 0;    
}

double Network::updateOneStep(int step){
    double average_square_queue = 0;
    // std::cout << "\n\tStep " << step << "\n";
    updateDepartures(step);
    for (unsigned int u = 0; u < lanes.size(); ++u){
        lanes[u].cleanUnprobableStates(step);
        lanes[u].updateQueues(step);
        if (TO_DELETE)
            average_square_queue += lanes[u].getAverageSquareAndRemoveQueues(step, getStartFrom(u, travel_times));
    }
    return average_square_queue;
}

double Network::run(int duration){
    double total_queue = 0;
    for (int step = starting_second; step < starting_second + duration; ++step){
        total_queue += updateOneStep(step);
    }
    if (!TO_DELETE)
        total_queue = getTotalQueue(duration);
    return total_queue;
}


double Network::getTotalQueue(int duration){
    double total = 0;
    int number_of_lanes = lanes.size();
    int number_of_queues = 0;
    for (int u = 0; u < number_of_lanes; ++u){
        const std::vector<double> & expectations = lanes[u].getNormalAverages(starting_second + duration);
        number_of_queues += lanes[u].getTotalNumberOfQueues();
        for (int i = starting_second; i < int(expectations.size()); ++i){
            total += expectations[i];
        }
    }
    return total;
}


double Network::getTotalSquareQueue(int duration){
    double total = 0;
    int number_of_lanes = lanes.size();
    int number_of_queues = 0;
    for (int u = 0; u < number_of_lanes; ++u){
        number_of_queues += lanes[u].getTotalNumberOfQueues();
        for (int i = starting_second; i < starting_second + duration; ++i){
            total += lanes[u].getAverageSquareQueue(i, getStartFrom(u, travel_times));
        }
    }
    // std::cout << "total number of queues is " << number_of_queues << "\n";
    return total;
}



void Network::printResultsToFile(int duration, const std::string & file_name){
    std::vector<double> results;
    int number_of_lanes = lanes.size();
    results.resize(duration * number_of_lanes);
    for (int u = 0; u < number_of_lanes; ++u){
        const std::vector<double> & expectations = lanes[u].getNormalAverages(duration);
        for (int i = 0; i < int(expectations.size()); ++i){
            results[i * number_of_lanes + u] = expectations[i];
        }
    }
    printVectorToFile(results, file_name);
}

void Network::printDistributionsToFile(int duration, const std::string & file_name){
    int number_of_lanes = lanes.size();
    std::vector<double> distribution;
    printStringToFile("[", file_name);
    for (int i = 0; i < duration; ++i){
        printStringToFile("[", file_name);
        for (int u = 0; u < number_of_lanes; ++u){
            distribution = lanes[u].getDistribution(i);
            printVectorToFile(distribution, file_name);
            if (u < number_of_lanes - 1)
                printStringToFile(", ", file_name);
        }
        printStringToFile("]", file_name);
        if (i < duration - 1)
            printStringToFile(", ", file_name);
    }
    printStringToFile("]", file_name);
}


void Network::printRightDistributionsToFile(int duration, const std::string & file_name){
    int number_of_lanes = lanes.size();
    std::vector<std::vector<std::vector<double>>> distributions;
    distributions.resize(number_of_lanes);
    for (int u = 0; u < number_of_lanes; ++u){
        distributions[u] = lanes[u].getRightCumulativeDistributions(starting_second, starting_second + duration);
    }
    duration = 120;
    for (int i = 0; i < duration; ++i){
        printStringToFile("[", file_name);
        for (int u = 0; u < number_of_lanes; ++u){
            printVectorToFile(distributions[u][i], file_name);
            if (u < number_of_lanes - 1)
                printStringToFile(", ", file_name);
        }
        printStringToFile("]", file_name);
        printStringToFile(", ", file_name);
    }
}
