#include <iostream>
#include <vector>
#include <set>
#include <cmath>
#include <algorithm>  
#include "general.h"    
#include "queues.h"
#include "state.h"
#include "rollqueues.h"
#include "deparrqueues.h"
#include "abpmap.h"


Queues::Queues(int duration, int capacity, std::map<char, ArrivalState> & a_states, Pair d_key) : capacity(capacity), untransitioned_queues(0, capacity, 0){
    queues.reserve(duration + 1);
    
    DepArrQueues first_queues;
    for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
        char a_key = getAKey(a_ind);
        if (isInMap(a_states, a_key)){
            double prob = a_states[a_key].getProbability();
            first_queues[d_key].emplace(a_ind, capacity, prob);        
        }
    }
    queues.push_back(first_queues);
    emptied_probs.resize(duration + 30);
}

Queues::Queues(int duration, int capacity, int travel_time, int short_travel_time, Pair first_d_key) : 
        capacity(capacity), untransitioned_queues(travel_time + 30, capacity, 0){
    queues.reserve(duration + travel_time + 30); // 30 for acceleration delay and other arrivals
    emptied_probs.resize(duration + travel_time + 30);
    inialiseInternalQueues(short_travel_time, first_d_key);
    
}

Queues::Queues(int duration, int capacity, int queue_length, Pair d_key, bool is_arrival, int starting_second) : 
        capacity(capacity), untransitioned_queues(0, capacity, 0){ // start from state queue_length for external lane
    queues.reserve(starting_second + duration + 1);
    DepArrQueues new_queues; 
        for (int i = 0; i < starting_second; ++i){
        addQueues(i, new_queues); // just empty queues
    }
    new_queues.emplace(d_key, is_arrival ? KEY_A : KEY_P, capacity, queue_length, 1.);        
    addQueues(starting_second, new_queues);
    emptied_probs.resize(starting_second + duration + 30);
}

Queues::Queues(int duration, int capacity, int travel_time, int queue_length, Pair first_d_key, 
               ArrivalChain & arrival_chain, DepartureChain & departure_chain, int starting_second) :
        capacity(capacity), untransitioned_queues(travel_time + 30, capacity, starting_second){ // start from state queue_length for internal lane
    queues.reserve(starting_second + duration + travel_time + 30); // 30 for acceleration delay and other arrivals
    emptied_probs.resize(starting_second + duration + travel_time + 30);
    inialiseInternalQueues(first_d_key, queue_length, arrival_chain, departure_chain, starting_second);
}

Queues::~Queues(){
    for (unsigned int i = 0; i < emptied_probs.size(); ++i){
        emptied_probs[i].clear();
    }
    emptied_probs.clear();
}



void Queues::inialiseInternalQueues(int travel_time, Pair first_d_key){
    DepArrQueues first_queues;
    first_queues.emplace(first_d_key, KEY_P, capacity, 1.); // = arr_queues;
    queues.push_back(first_queues);
    untransitioned_queues.addQueues(0, first_queues);
    first_queues.eraseAllQueues();
    for (int i = 1; i < travel_time; ++i){
        queues.push_back(first_queues);
        untransitioned_queues.addQueues(i, first_queues);
    }
}


void Queues::inialiseInternalQueues(Pair first_d_key, int queue_length, ArrivalChain & arrival_chain, DepartureChain & departure_chain, 
                                    int starting_second){
    DepArrQueues new_queues; 
    std::map<char, std::map<char, double>> ar_transitions_to;
    double unblocked;
    char a_key = arrival_chain.getStates(starting_second).begin()->first;
    
    for (int i = 0; i < starting_second; ++i){
        addQueues(i, new_queues); // just empty queues
    }
    
    new_queues.emplace(first_d_key, getInd(a_key), capacity, queue_length, 1.); // = arr_queues;
    addQueues(starting_second, new_queues);
    untransitioned_queues.addDownstreamPartsOfQueues(starting_second, starting_second - 1, new_queues);
    new_queues.removeDownstreamParts(starting_second, starting_second - 1);
    for (int i = starting_second; i < int(arrival_chain.getSize()); ++i){
        if (i < int(arrival_chain.getSize() - 1)) {
            // in this case the transitions are not zero
            ar_transitions_to = arrival_chain.getNormalisedTransitionsToKeys(i); //getAllArTransitionsToKeys(second, arrival_chain);
            new_queues = updateQueuesOneStep(i, unblocked, starting_second - 1, capacity, new_queues, ar_transitions_to, ar_transitions_to, 
                                                arrival_chain, departure_chain, emptied_probs); 
            addQueues(i + 1, new_queues);
            // add to untransitioned queues at second because these are queues multiplied by probability that there was no transition
            untransitioned_queues.addDownstreamPartsOfQueues(i + 1, starting_second - 1, new_queues);
            // remove downstream part
            new_queues.removeDownstreamParts(i + 1, starting_second - 1);
        } else {
            // transitions are zero just everything to untransitioned
            untransitioned_queues.addQueues(i, new_queues);
        }
    }
}


void Queues::printQueues(int second){
    std::cout << "Printing queues for second " << second << "\n";
    for (auto & pair_a_queues: queues[second].queues){
        std::cout << pair_a_queues.first << ": \n";
        for (int i = 0; i < NUMBER_OF_KEYS; ++ i){
            if (pair_a_queues.second.hasKey(i)){
                std::cout << i << " ";
                pair_a_queues.second[i].printQueue();
            }
        }
    }
    double total = queues[second].getTotalProbability();
    std::cout << total << " (" << 1 - total << ")\n";
    std::cout << "finish printing queues\n";
}

void Queues::printKeys(int second){
    std::cout << "Printing queue keys for second " << second << "\n";
    for (auto & pair_a_queues: queues[second].queues){
        std::cout << pair_a_queues.first << ": \n";
        for (int i = 0; i < NUMBER_OF_KEYS; ++ i){
            if (pair_a_queues.second.hasKey(i))
                std::cout << i << ", ";
        }
        std::cout << "\n";
    }
    std::cout << "finish printing keys\n";
}



void Queues::updateQueues(int step, ArrivalChain& arrival_chain, DepartureChain& departure_chain){
    DepArrQueues new_queues; 
    std::map<char, std::map<char, double>> normalised_to_keys_transitions = arrival_chain.getNormalisedTransitionsToKeys(step);
    last_ar_transitions = flipKeys(normalised_to_keys_transitions);
    double unblocked;
    queues.push_back(updateQueuesOneStep(step, unblocked, step, capacity, queues[step], normalised_to_keys_transitions, 
                                                                  normalised_to_keys_transitions, arrival_chain, departure_chain, emptied_probs));
    assert(checkProbabilities(step, arrival_chain));
    deleteEmptyQueues(step + 1); // empty in the sense of start = end
}

void Queues::updateLastSecond ( int step, int travel_time, ArrivalChain& arrival_chain, DepartureChain& departure_chain ){
    /*Update the transition probabilities for second  step + travel_time - 1  and queue probabilities for second  step + travel_time.*/
    std::map<char, double> coefs;
    std::map<char, double> coefs_full;
    std::map<char, std::map<char, double>> ar_transitions_to; // , ar_transitions_to_full;
    int travel_time_second = step + travel_time - 1;
    DepArrQueues new_queues; 
    Pair d_key;
    double unblocked = 0;
    
    // compute default transitions
    for (auto & a_key: {'a', 'p', 'b'}){
        if (arrival_chain.hasState(travel_time_second, a_key)){
            ar_transitions_to[default_ar_transitions.at(a_key)][a_key] = 1;  
        }
    }
    // get queues to update (and delete them from the untransitioned queues)
    int m_second = minNotAffectedQueueLength(travel_time_second, step - 1, capacity);
    for (auto & d_key_queues: untransitioned_queues[travel_time_second].queues){
        d_key = d_key_queues.first;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues.second.hasKey(a_ind) && d_key_queues.second[a_ind].getEnd() > m_second && 
                    d_key_queues.second[a_ind].isPartialProbabilitySignificant(m_second, capacity + 1)){ // upstream side exists and not zero
                new_queues[d_key].insert(a_ind, d_key_queues.second[a_ind].getUpstreamSide(travel_time_second, step, true)); // and delete
            }
        }
    }
    // update new queues according to the default transitions
    // no need to delete as we added only non-zero queues
    // TODO: what if travel time is 1? shouldn't real emptied_probs be used?
    new_queues = updateQueuesOneStep(travel_time_second, unblocked, step, capacity, new_queues, 
                                     ar_transitions_to, ar_transitions_to, arrival_chain, departure_chain, emptied_probs);
    // add new_queues to queues and to untransitioned queues
    addQueues(travel_time_second + 1, new_queues);
    untransitioned_queues.addQueues(travel_time_second + 1, new_queues);
    // add default transitions to the arrival_chain
    arrival_chain.addDefaultTransitions(travel_time_second, unblocked);
    
}


void Queues::checkCapacities(){
    int queue_capacity;
    for (int i = 0; i < int(queues.size()); ++i){
        for (auto & d_iter: queues[i].queues){
            for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
                if (d_iter.second.hasKey(a_ind)){
                    queue_capacity = d_iter.second[a_ind].getCapacity();
                    if (queue_capacity != capacity)
                        std::cout << "Wrong capacity: " << i << " " << d_iter.first << " " << a_ind << " " << queue_capacity << "\n";
                }
            }
        }
    }
}


bool Queues::checkProbabilities(int second, const ArrivalChain& arrival_chain){
    std::map<int, double> total_probabilities;
    for (auto & d_iter: queues[second].queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_iter.second.hasKey(a_ind)){
                addValue(total_probabilities, a_ind, d_iter.second[a_ind].getTotalProbability());
            }
        }
    }
    for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
        char a_key = getAKey(a_ind);
        if (arrival_chain.hasState(second, a_key)){
            double prob = arrival_chain.getStateProbability(second, a_key);
            if (!almostEqual(total_probabilities[a_ind], prob, big_eps)){
                printQueues(second-1);
                printQueues(second);
                arrival_chain.printStates(second);
                std::cout << a_key << " " << total_probabilities[a_ind] - prob << "\n";
                std::cout << "Queue probabilities and arrival-chain-states probabilities are not the same!\n";
                return false;
            }
        }
    }
    return true;
}


void Queues::checkPartialProbabilities(int second, ArrivalChain& arrival_chain){
    if (second < int(queues.size())){
        std::map<int, double> total_probabilities;
        for (auto & d_iter: queues[second].queues){
            for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
                if (d_iter.second.hasKey(a_ind)){
                    addValue(total_probabilities, a_ind, d_iter.second[a_ind].getTotalProbability());
                }
            }
        }
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            char a_key = getAKey(a_ind);
            if (arrival_chain.hasState(second, a_key)){
                double prob = arrival_chain.getStateProbability(second, a_key);
                if (total_probabilities[a_ind] > prob + big_eps){
                    printQueues(second-1);
                    printQueues(second);
                    arrival_chain.printStates(second-1);
                    arrival_chain.printStates(second);
                    std::cout << a_key << " " << total_probabilities[a_ind] - prob << "\n";
                    switchStyle("red");
                    std::cout << "Queue probabilities are larger than corresponding arrival-chain-states probabilities!\n";
                    switchStyle("reset");
                    throw "Queue probabilities are larger than corresponding arrival-chain-states probabilities!\n";
                }
            }
        }
    }
}


std::map<char, std::map<char, double>> flipKeys(const std::map<char, std::map<char, double>> & transitions){
    std::map<char, std::map<char, double>> new_transitions;
    for (auto & key1_iter: transitions){
        for (auto & key2_iter: key1_iter.second){
            new_transitions[key2_iter.first][key1_iter.first] = key2_iter.second;
        }
    }
    return new_transitions;
}

void Queues::updateInternalQueues (int step, int travel_time, ArrivalChain& arrival_chain, DepartureChain& departure_chain ){
    // update arrival flows for step + travel_time - 1
        
    updateLastSecond(step, travel_time, arrival_chain, departure_chain);
    assert(arrival_chain.checkTotalTransitionProbability(step + travel_time - 1));
    DepArrQueues new_queues; 
    last_ar_transitions = flipKeys(arrival_chain.getNormalisedTransitionsToKeys(step, step - 1, *this));
    for (int i = step; i < int(arrival_chain.getSize()); ++i){
        updateDeparturesAndQueuesOneStep(step, i, new_queues, arrival_chain, departure_chain);
        assert(untransitioned_queues.checkAllUntransitioned(step, arrival_chain));
    }
    deleteEmptyQueues(step + 1);
    untransitioned_queues.deleteOldQueues();
    
    //run all checks
    assert(arrival_chain.checkTotalTransitionProbability(step + travel_time - 1));
    assert(arrival_chain.checkTotalProbability(step + travel_time));
    assert(arrival_chain.checkTransitionedFlows(step));
    assert(checkAndPrintTotalProbability(step + 1, true, false));
    
    assert(untransitioned_queues.checkTotalProbability());
    assert(untransitioned_queues.checkUntransitioned(step + 1, arrival_chain));
}

std::map<char, double> getUntransitionedFromToKeys(std::map<char, std::map<char, double>> transitions_to_keys){
    std::map<char, double> result;
    for (auto & a_key_to_iter: transitions_to_keys){
        for (auto & a_key_from_iter: a_key_to_iter.second){
            addValue(result, a_key_from_iter.first, -a_key_from_iter.second);
        }
    }
    for (auto & a_key: {'a', 'p', 'b'}){
        addValue(result, a_key, 1.);
    }
    return result;
}

void Queues::updateDeparturesAndQueuesOneStep ( int step, int second, DepArrQueues& new_queues, 
                                                ArrivalChain& arrival_chain, DepartureChain& departure_chain){
    DepArrQueues middle_queues;
    Pair d_key;
    
    int m_down = minNotAffectedQueueLength(second, step, capacity); // middle is inbetween
    int m_up = minNotAffectedQueueLength(second, step-1, capacity);
    
    new_queues.deleteEmptyQueues();
    if (0 < m_up){
        for (auto & d_key_queues: queues[second].queues){
            d_key = d_key_queues.first;
            for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
                if (d_key_queues.second.hasKey(a_ind)){
                    Queue & queue = d_key_queues.second[a_ind];
                    int new_start, new_end;
                    queue.getStartEndMiddle(0, m_up, new_start, new_end);
                    if (new_start < new_end && queue.isPartialProbabilitySignificant(new_start, new_end)){
                        middle_queues[d_key].emplace(a_ind, capacity, new_start, new_end); // this is always possible because new_start < new_end and capacity is ok
                        swapQueueParts(middle_queues[d_key][a_ind], queue, new_start, new_end);
                        queue.removeMiddlePartKnowingMs(0, m_up, false);
                        untransitioned_queues.removeMiddlePart(second, d_key, a_ind, 0, m_up);
                    }
                }
            }
        }
    }
    
    updateDepartures(second - 1, step, capacity, middle_queues, departure_chain, emptied_probs[second]);
 
    
    addQueues(second, middle_queues);
    untransitioned_queues.addDownstreamPartsOfQueues(second, step, middle_queues);
    middle_queues.removeDownstreamParts(m_down);
    
    new_queues.addQueues(middle_queues);
    // remove downstream part
    new_queues.removeDownstreamParts(second, step);
    
    //update arrivals and departures
    std::map<char, std::map<char, double>> ar_transitions_to, ar_transitions_to_full;
    ar_transitions_to = arrival_chain.getNormalisedTransitionsToKeys(second, step - 1, *this); //getAllArTransitionsToKeys(second, arrival_chain);
    // here step - 1 is used since this is how m_up is defined
    ar_transitions_to_full = arrival_chain.getNormalisedFullTransitionsToKeys(second, *this); //getAllArTransitionsFullToKeys(second, arrival_chain);
    std::map<char, double> untransitioned_probs = getUntransitionedFromToKeys(ar_transitions_to);
    std::map<char, double> untransitioned_full_probs = getUntransitionedFromToKeys(ar_transitions_to_full); 
    
    untransitioned_queues.addMiddleAndUpstreamPartsWithCoef(second, step, new_queues, untransitioned_probs, untransitioned_full_probs); 
    double unblocked;
    // update is needed even in case of last second, because we check total sum
    new_queues.deleteEmptyQueues();
    new_queues = updateQueuesOneStep(second, unblocked, step, capacity, new_queues, ar_transitions_to, ar_transitions_to_full, 
                                     arrival_chain, departure_chain, emptied_probs); 
    if (second + 1 < arrival_chain.getSize()){
        
        addQueues(second + 1, new_queues);
        // add to untransitioned queues at second because these are queues multiplied by probability that there was no transition
        untransitioned_queues.addDownstreamPartsOfQueues(second + 1, step - 1, new_queues);
        // remove downstream part
        new_queues.removeDownstreamParts(second + 1, step - 1);
    }
}

std::map<char, double> Queues::getArTransitions ( int second, char a_key, ArrivalChain& arrival_chain ){
    std::map<char, double> transitions;
    if (a_key == 'a'){
        transitions = {{'p', 1}};
    } else if (a_key == 'p'){
        double transition_flow = arrival_chain.getTransitionFlow(second, a_key, 'a');
        double not_full = arrival_chain.getStateProbability(second, a_key) - getAlmostFullProbability(second, a_key);
        if (not_full > almost_zero){
            if (transition_flow > 0) {
                transitions['a'] = transition_flow / not_full;
                if (transition_flow < not_full){
                    transitions['p'] = 1 - transitions['a'];
                }
            } else {
                transitions['p'] = 1;
            }
        }
    } // if 'b' then there are no transitions as the queue should be full
    return transitions;
}

std::map<char, std::map<char, double> > Queues::getAllArTransitions ( int second, ArrivalChain& arrival_chain ){
    std::map<char, std::map<char, double>> results;
    if (arrival_chain.hasState(second, 'a')){
        results['a'] = getArTransitions(second, 'a', arrival_chain);
    }
    if (arrival_chain.hasState(second, 'p')){
        results['p'] = getArTransitions(second, 'p', arrival_chain);
    }
    return results;
}

std::map<char, double> Queues::getArTransitionsFull ( int second, char a_key, ArrivalChain& arrival_chain ){
    std::map<char, double> transitions;
    double transition_flow = arrival_chain.getTransitionFlow(second, a_key, 'b');
    double full = getAlmostFullProbability(second, a_key);
    if (full > almost_zero){
        if (transition_flow > eps) {
            transitions['b'] = transition_flow / full;
            if (transition_flow < full){
                transitions['p'] = 1 - transitions['b'];
            }
        } else {
            transitions['p'] = 1;
        }
    }
    return transitions;
}

std::map<char, std::map<char, double> > Queues::getAllArTransitionsFull ( int second, ArrivalChain& arrival_chain ){
    std::map<char, std::map<char, double>> results;
    for (auto & a_key: {'b', 'a', 'p'}){
        if (arrival_chain.hasState(second, a_key)){
            results[a_key] = getArTransitionsFull(second, a_key, arrival_chain);
        }
    }
    return results;
}

std::map<char, std::map<char, double> > Queues::getAllArTransitionsFullToKeys(int second, ArrivalChain& arrival_chain){
    std::map<char, std::map<char, double>> results_from, results_to;
    results_from = getAllArTransitionsFull(second, arrival_chain);
    for (auto & a_iter_from: results_from){
        for (auto & a_iter_to: a_iter_from.second){
            results_to[a_iter_to.first][a_iter_from.first] = a_iter_to.second;
        }
    }
    return results_to;
}

std::map<char, std::map<char, double> > Queues::getAllArTransitionsToKeys(int second, ArrivalChain& arrival_chain){
    std::map<char, std::map<char, double>> results_from, results_to;
    results_from = getAllArTransitions(second, arrival_chain);
    for (auto & a_iter_from: results_from){
        for (auto & a_iter_to: a_iter_from.second){
            results_to[a_iter_to.first][a_iter_from.first] = a_iter_to.second;
        }
    }
    return results_to;
}


std::vector<double> Queues::getNormalAverages(int duration){
    std::vector<double> results(duration, 0);
    int start, end, new_second;
    double total, prob;
    for (int second = 0; second < duration; ++second){
//         std::cout << queues[second].queues.size() << "\n";
        for(auto & d_iter: queues[second].queues){
//             std::cout << d_iter.first << "\n";
            for(int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
                if (d_iter.second.hasKey(a_ind)){
                    d_iter.second[a_ind].getStartEnd(start, end);
                    total = 0;
                    for (int i = end - 1; i >= 1; --i){
                        prob = d_iter.second[a_ind][i];
                        total += prob;
                        new_second = int(second - (i - 1) * bumper_to_bumper / free_flow_speed);
                        if (new_second >= 0)
                            results[new_second] += total;
                    }
                }
            }
        }
    }
    return results;
}



bool Queues::checkAndPrintTotalProbability(int second, bool throw_exception /*= true*/, bool print /*= true*/){
    double total, prob;
    std::map<char, double> per_a_key;
    total = queues[second].getTotalProbability();
    for (auto & d_key_queues: queues[second].queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues.second.hasKey(a_ind)){
                prob = d_key_queues.second[a_ind].getTotalProbability();
                addValue(per_a_key, getAKey(a_ind), prob);
            }
        }
    }
    if (print){
        std::cout << "Total probability is " << total << "\n";
        std::cout << "per a keys: " << per_a_key << "\n";
        std::cout << 1 - total << "\n";
    }
    if ((total + eps < 1) && throw_exception){
        printQueues(second - 1);
        checkAndPrintTotalProbability(second, false, true);
        printQueues(second);
        std::cout << " Too low total probability!\n ";
        return false;
    } else if (total > 1 + eps){
        printQueues(second - 1);
        checkAndPrintTotalProbability(second - 1, false, true);
        printQueues(second);
        std::cout << "Total probability is " << total << "\n";
        std::cout << 1 - total << "\n";
        std::cout << " Too high total probability!\n ";
        return false;
    } 
    return true;
}


void Queues::renormaliseQueues(int second, double total){
    queues[second].renormaliseQueues(total);
}


void printDepArrQueues(const DepArrQueues & queues){
    std::cout << "Print dep arr queues\n";
    queues.printQueues();
}

double getTotalProbabilityDepArrQueues(const DepArrQueues & queues){
    return queues.getTotalProbability();
}

double printTotalProbabilityDepArrQueues(const DepArrQueues & queues){
    double total; // = 0., prob;
    total = getTotalProbabilityDepArrQueues(queues);
    std::cout << "Total probability is " << total << "\n";
    return total;
}


void Queues::checkProblemQueue(int problem_index){
    switchStyle("cyan");
    if (problem_index < int(queues.size())){
        printQueues(problem_index - 1);
        checkAndPrintTotalProbability(problem_index - 1, false);
        printQueues(problem_index);
        std::cout << "Untransitioned:\n";
        untransitioned_queues.printQueues(problem_index - 1, false);
        untransitioned_queues.printQueues(problem_index, false);
        checkAndPrintTotalProbability(problem_index, false);
    }
    switchStyle("reset");
}


void Queues::deleteEmptyQueues(int second){
    queues[second].deleteEmptyQueues();
}

double Queues::getAverageQueue(int second){
    return queues[second].getAverageQueue();
}

double Queues::getAverageQueueInMeters(int second){
    double result = 0.;
    for (auto & d_key_iter: queues[second].queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_iter.second.hasKey(a_ind))
                result += d_key_iter.second[a_ind].getExpectedQueueLengthInMeters();
        }
    }
    return result;
}


std::map<KeyDA, double> Queues::getZeroProbabilities(int second){
    // NOTE: only states with probably non-zero zero probability are included
    std::map<KeyDA, double> results;
    for (auto & d_key_iter: queues[second].queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_iter.second.hasKey(a_ind) && d_key_iter.second[a_ind].getStart() == 0)
                results[KeyDA(d_key_iter.first, getAKey(a_ind))] = d_key_iter.second[a_ind][0];
        }
    }
    return results;
}

double Queues::getTotalUntransitionedQueue(int second, int downstream_second, char a_key){
    return untransitioned_queues.getTotalUntransitionedQueue(second, downstream_second, a_key);
}


double Queues::getAlmostFullProbability(int second, char a_key){
    double total = 0;
    int which = (a_key == 'a') ? capacity - 1: capacity;
    for (auto & d_key_iter: queues[second].queues){
        if (d_key_iter.second.hasKey(a_key)){
            total += d_key_iter.second[a_key][which];
        }
    }
    return total;
}

double Queues::getAlmostFullUntransitionedProbability(int second, char a_key){
    return untransitioned_queues.getAlmostFullProbability(second, a_key);
}


std::map<KeyIA, std::map<Pair, double> > Queues::getAlmostFullDKeys(int second){
    std::map<KeyIA, std::map<Pair, double>> results;
    std::map<Pair, double> inter_results;
    inter_results = untransitioned_queues.getAlmostFullDKeys(second, 'p');
    if (inter_results.size() > 0)
        results[KeyIA(second, 'p')] = inter_results;
    for (unsigned int i = second; i < queues.size(); ++i){
        inter_results = untransitioned_queues.getAlmostFullDKeys(i, 'a');
        if (inter_results.size() > 0)
            results[KeyIA(i, 'a')] = inter_results;
        inter_results = untransitioned_queues.getAlmostFullDKeys(i, 'b');
        if (inter_results.size() > 0)
            results[KeyIA(i, 'b')] = inter_results;
    }
    return results;
}


bool checkTotalNewProbability(const DepArrQueues & new_queues, const DepArrQueues & queues, double prob_will_be, 
                              int second, int downstream_second, const std::map<char, std::map<char, double>>& ar_transitions_to, 
                              const std::map<char, std::map<char, double>>& ar_transitions_full_to){
    double result = new_queues.getTotalProbability();
    if ((result + small_eps < prob_will_be) || (result > prob_will_be + small_eps)){
        std::cout << second << " " << downstream_second << "\n";
        printDepArrQueues(queues);
        printDepArrQueues(new_queues);
        std::cout << ar_transitions_to << "\n";
        std::cout << ar_transitions_full_to << "\n";
        std::cout << "The queue probability changed incorrectly!\n";
        return false;
    }
    return true;
}

bool areTransitionsEmpty(std::map<char, std::map<char, double>>& ar_transitions_to, std::map<char, std::map<char, double>>& ar_transitions_full_to){
    return (ar_transitions_full_to.size() == 0) && (ar_transitions_to.size() == 0);
}


DepArrQueues updateQueuesOneStep(int second, double & unblocked, int downstream_second, int capacity, const DepArrQueues & queues, 
                                 std::map<char, std::map<char, double>>& ar_transitions_to, 
                                 std::map<char, std::map<char, double>>& ar_transitions_full_to, 
                                 ArrivalChain & arrival_chain, DepartureChain& departure_chain, std::vector<std::map<KeyDA, double>> & emptied){
    // Update queues for second.
    // Note: arrival transitions should be ordered with [a_key_to][a_key_from]!
    DepArrQueues new_queues;
    if (areTransitionsEmpty(ar_transitions_to, ar_transitions_full_to) || queues.queues.size() == 0)
        return new_queues; // no computations are needed
    Pair d_key;
    double prob_will_be = 0, prob;  
    std::map<int, double> probs_not_full = {{'a', 0}, {'p', 0}, {'b', 0}};
    std::map<int, double> probs_full = {{'a', 0}, {'p', 0}, {'b', 0}};
    double almost_full;
    for (auto & d_key_iter: queues.queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_iter.second.hasKey(a_ind)){
                almost_full = d_key_iter.second[a_ind].getAlmostFullProbability(a_ind == KEY_A);
                probs_not_full[a_ind] += d_key_iter.second[a_ind].getTotalProbability() - almost_full;
                probs_full[a_ind] += almost_full;
            }
        }
    }

    MapCharChar transitioned, transitioned_full;
    for (auto & a_iter_to: ar_transitions_to){
        for (auto & a_iter_from: a_iter_to.second){
            prob = probs_not_full[getInd(a_iter_from.first)] * a_iter_from.second;
            prob_will_be += prob;
            transitioned[a_iter_from.first][a_iter_to.first] = prob;
        }
    }
    for (auto & a_iter_to: ar_transitions_full_to){
        for (auto & a_iter_from: a_iter_to.second){
            prob = probs_full[getInd(a_iter_from.first)] * a_iter_from.second;
            prob_will_be += prob;
            transitioned_full[a_iter_from.first][a_iter_to.first] = prob;
        }
    }
    
    // Process arrivals.
    for (auto & d_key_iter: queues.queues){
        d_key = d_key_iter.first;
        bool is_free = departure_chain.isFreeDepartureAllowed(d_key, second);
        for (auto & a_key_to: {'a', 'b', 'p'}){
            if (isInMap(ar_transitions_to, a_key_to) || isInMap(ar_transitions_full_to, a_key_to))
                new_queues.emplace(d_key, getInd(a_key_to), capacity, d_key_iter.second.queues, ar_transitions_to[a_key_to], ar_transitions_full_to[a_key_to], is_free);
        }
    }
    if (int(emptied.size()) <= second + 1){
        std::cout << emptied.size() << "\n";
        std::cout << second << "\n";
        throw "Incorrect size of emptied vector\n";
    }
    unblocked = updateDepartures(second, downstream_second, capacity, new_queues, departure_chain, emptied[second + 1]);
    assert(checkTotalNewProbability(new_queues, queues, prob_will_be, second, downstream_second, ar_transitions_to, ar_transitions_full_to));
    new_queues.deleteEmptyQueues();
    arrival_chain.addTransitionedProbabilities(second, transitioned, transitioned_full, unblocked);
    return new_queues; 
}

 
void Queues::updateDepartureState(int step, const std::map<KeyDA, std::map<Pair, double> >& tr_flows_zero){
    // update departure state for B_step = 0, L_step = a, R_step = d to X_step = (0, a, d') with probability tr_probabilities_zero[d, a][d']
    for (auto & da_iter: tr_flows_zero){
        Pair d_key = da_iter.first.first;
        char a_key = da_iter.first.second;
        double total = 0;
        if (!isInMap(queues[step].queues, d_key) || !queues[step][d_key].hasKey(a_key)){
            total += sumMapValues(da_iter.second);
            if (total < small_eps / 1000){
                // not bad, but don't do anything
                continue;
            } else {
                // bad, throw exception
                std::cout << total << "\n";
                std::cout << da_iter.first << "\n";
                std::cout << tr_flows_zero << "\n";
                printQueues(step);
                throw "Flow transitioned from a state that does not exists in the queues!\n";
            }
        }
        total = 0;
        for (auto & d_iter: da_iter.second){
            Pair unblock_d_key = d_iter.first;
            double prob = d_iter.second;
            total += prob;
            if (isInMap(queues[step].queues, unblock_d_key) &&  queues[step][unblock_d_key].hasKey(a_key)){
                queues[step][unblock_d_key][a_key].addProbability(0, prob);
            } else {
                queues[step].emplace(unblock_d_key, getInd(a_key), capacity, prob);
            }
        }
        queues[step][d_key][a_key].addProbability(0, - total);
    }
}
 
bool checkTailProbability(const DepArrQueues & new_queues, double tail_prob){
    double total = new_queues.getTotalProbability();
    if (total > tail_prob + eps){
        new_queues.printQueues();
        std::cout << "Tail is thinner than the obtained queues probabilities.";
        return false;
    }
    return true;
}


void createTailQueue(DepArrQueues & new_queues, Pair & d_key, char a_key_from, char a_key_to, ArrQueues & arr_queues, int m_second, double coef){
    if (arr_queues.hasKey(a_key_from)){  // && arr_queues[a_key_from].getEnd() > m_second && arr_queues[a_key_from].getPartialProbability(m_second) > almost_zero){            
        if (new_queues.emplace(d_key, getInd(a_key_from), arr_queues[a_key_from], m_second, (a_key_from == 'a'), (a_key_to == 'b'), coef)){
            int new_start = 0, new_end = 0;
            new_queues[d_key][a_key_from].getStartEnd(new_start, new_end);
            arr_queues[a_key_from].removeTailPart(new_start, new_end, coef);
        }
    }
}
 
double Queues::processTail (int step, const Tail& tail, ArrivalChain& arrival_chain, DepartureChain& departure_chain ){
    double unblocked = 0;
    int second_from = tail.second_from, second_to = tail.second_to;
    double tail_prob = tail.probability, coef, divide_by;
    char a_key_from = tail.a_key_from, a_key_to = tail.a_key_to;
    DepArrQueues new_queues;
    Pair d_key;
    if (tail_prob < - small_eps)
        throw "Negative tail probability!\n";
    if (a_key_from == 'b')
        throw "Tail attached to b!\n";
    if (tail_prob < almost_zero)
        return unblocked;
    #ifndef NDEBUG
        double were = 0;
        if ((a_key_to == 'b') && (int(queues.size()) > second_to)){
            were = getAlmostFullProbability(second_to, 'b');
        }
    #endif
    if (!arrival_chain.hasState(second_from, a_key_from) && tail_prob < almost_zero * 100){
        return unblocked;  // do not update such small tails
    }
    double almost_full = untransitioned_queues.getAlmostFullProbability(second_from, a_key_from);
    if (a_key_to == 'a'){
        divide_by = arrival_chain.getUnknownTransitionProbability(second_from, a_key_from) - almost_full; 
    } else { // 'b'
        divide_by = almost_full;
    }
    if (divide_by > almost_zero)
        coef = tail_prob / divide_by;
    else 
        coef = 0;
    if (coef > 1 + eps && divide_by * coef > small_eps + divide_by){
        arrival_chain.printStates(second_from);
        std::cout << coef << " " << coef -1 << "\n";
        std::cout << coef << " " << divide_by << " " << almost_full << "\n";
        throw "Big coef!\n";
    }
    coef = std::min(coef, 1.);
    
    int m_second = minNotAffectedQueueLength(second_from, step - 1, capacity);
    for (auto & d_iter: untransitioned_queues[second_from].queues){
        d_key = d_iter.first;
        createTailQueue(new_queues, d_key, a_key_from, a_key_to, d_iter.second, m_second, coef);
    }
    assert(checkTailProbability(new_queues, tail_prob));
    
    for (int i = second_from; i < second_to; ++i){
        
        std::map<char, std::map<char, double>> ar_transitions_to;
        ar_transitions_to[(i == second_to - 1) ? a_key_to: 'p'] = {{(i == second_from) ? a_key_from : 'p', 1}};
        new_queues = updateQueuesOneStep(i, unblocked, step - 1, capacity, new_queues, ar_transitions_to, ar_transitions_to, arrival_chain, 
                                                                    departure_chain, emptied_probs);
        addQueues(i + 1, new_queues);
        untransitioned_queues.addDownstreamPartsOfQueues(i + 1, step - 1, new_queues); // step -1 so as to add middle queues as well
        
        new_queues.removeDownstreamParts(i + 1, step - 1);
        new_queues.deleteEmptyQueues();
        if (unblocked > tail_prob + small_eps){
            throw "Unblocked more than was possible for this tail.\n";
        }
        unblocked = std::min(unblocked, tail_prob);
        if (a_key_to != 'b') {
            arrival_chain.addFlow(i, (i == second_from) ? a_key_from : 'p', (i == second_to - 1) ? a_key_to: 'p', tail_prob);
        } else {
            arrival_chain.addFlow(i, a_key_from, 'b', tail_prob - unblocked);
            if (unblocked > 0)
                arrival_chain.addFlow(i, a_key_from, 'p', unblocked);
        }
    }
    untransitioned_queues.addQueues(second_to, new_queues);
    #ifndef NDEBUG
    if (a_key_to == 'b'){
        double became = getAlmostFullProbability(second_to, 'b');
        if ((became + eps + unblocked < were + tail_prob) || (became + unblocked > were + tail_prob + eps)){
            throw "While adding blocking tail, the probability was not changed accordingly\n";
        }
    }
    #endif
    return unblocked;
}

void Queues::addQueues ( int second, DepArrQueues& new_queues ){
    new_queues.deleteEmptyQueues();
    if (int(queues.size()) <= second){
        //no such second yet, just push back
        queues.push_back(new_queues);
        if (int(queues.size()) <= second){
            throw "Wrong queues size while adding new queues\n";
        }
    } else { // add to existing second
        queues[second].addQueues(new_queues);
    }
}


std::map<KeyDA, double> Queues::getJustEmptiedProbabilities(int step, bool is_split_lane, bool is_internal) {
    if (is_split_lane){ // transfer zero-queue-length probability to state (0, 0)
        queues[step].redirectZeroProbabilityTo(Pair(0, 0), capacity);
        if (is_internal)
            untransitioned_queues.redirectZeroProbabilityTo(step, Pair(0, 0));
    }
    return emptied_probs[step];
}



bool checkProbabilityAfterDepartures(const DepArrQueues & queues, double was, const DepArrQueues & were_queues, int second, int downstream_second){
    double became = getTotalProbabilityDepArrQueues(queues);
    if ((became + small_eps < was) || (became > was + small_eps)){
        std::cout << "were:\n";
        were_queues.printQueues();
        std::cout << "became:\n";
        queues.printQueues();
        std::cout << second << " " << downstream_second << "\n";
        std::cout << was << " -> " << became << "\n";
        std::cout << became - was << "\n";
        std::cout << "Changed total probability after departures\n";
        return false;
    }
    return true;
}
 
double updateDepartures(int second, int downstream_second, int capacity, DepArrQueues &queues, DepartureChain& departure_chain, 
                        std::map<KeyDA, double> & emptied){ 
    // process departures
    double unblocked = 0;
    if (queues.queues.size() == 0)
        return unblocked;
    
    #ifndef NDEBUG
        DepArrQueues were_queues;
        double was;
        was = getTotalProbabilityDepArrQueues(queues);
        were_queues = queues;
    #endif
    
    std::set<Pair> d_keys; // (ordered) set of all d_keys 
    for (auto const & d_key_queues: queues.queues){
        Pair d_key = d_key_queues.first;
        d_keys.insert(d_key);
    }
    
    for (auto const & d_key: d_keys){
        int last = departure_chain.getLastSecond(d_key);
        if (last <= downstream_second){
//             int m_second;  // second + 1 - (m_second-1)^+ * bumper_to_bumper * (1/free_flow_speed + 1/ wave_speed) - bumper_to_bumper/free_flow_speed <= last
            int m_second = minNotAffectedQueueLength(second + 1, last, capacity); // second + 1 is used as the departures are for the next (!) second
            std::map<Pair, double> transitions = departure_chain.getTransitions(d_key); // \phi^{dd'}_+
            std::map<Pair, double> departure_probabilities = departure_chain.getDepartureProbabilities(d_key); // \mu^{dd'}

            for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
                if (!queues[d_key].hasKey(a_ind) || queues[d_key][a_ind].getStart() >= m_second)
                    continue;
                double zero_prob = queues[d_key][a_ind][0]; // this should be transitioned to g(d_key)
                queues[d_key][a_ind].setProbability(0, 0);
                for (auto const & transtion_iter: transitions){
                    Pair d_key_to = transtion_iter.first;
                    double tr_prob = transtion_iter.second;
                    double dep_prob = departure_probabilities[d_key_to];
                    if (tr_prob > 0.){
                        d_keys.insert(d_key_to);  // insert so that it can be processed later (by definition, it should be that d_key_to > d_key)
                        if (queues.queues.find(d_key_to) == queues.queues.end() || 
                                !queues[d_key_to].hasKey(a_ind)){
                            // no such queue yet
                            queues.emplace(d_key_to, a_ind, capacity, queues[d_key][a_ind], m_second, 
                                                     (1 - dep_prob) * tr_prob, dep_prob * tr_prob/*, is_split_lane*/);
                        } else { // there is such queue
                            queues[d_key_to][a_ind].addDepartureQueue(queues[d_key][a_ind], m_second, (1 - dep_prob) * tr_prob, dep_prob * tr_prob/*, is_split_lane*/);

                        }
                        double emptied_prob = queues[d_key][a_ind][1] * dep_prob * tr_prob;
                        if (emptied_prob > almost_zero){
                            // it is important to know which d_key_to is emptied even if the transition is to (0,0)
                            addValue(emptied, KeyDA(d_key_to, getAKey(a_ind)), emptied_prob); 
                        }                        
                    }
                }
                
                if (queues[d_key][a_ind].removeDownstreamPart(m_second)){ // queue is empty after removing downstream part
                    queues[d_key].eraseItem(a_ind);
                    if (!queues[d_key].hasAnyKey())
                        queues.eraseItem(d_key);
                }
                
                // add default transition d -> g(d) for m = 0
                if (zero_prob > 0) {
                    Pair d_key_to = departure_chain.getDefaultZeroTransition(d_key); // default transition of d_key 
//                  // NOTE: here the transition is different than the default transition!!!
                    d_keys.insert(d_key_to); 
                    if (queues.queues.find(d_key_to) == queues.queues.end() || 
                        !queues[d_key_to].hasKey(a_ind)){
                                // no such queue yet
                        queues.emplace(d_key_to, a_ind, capacity, zero_prob);
                    } else { // there is such queue
                        queues[d_key_to][a_ind].addProbability(0, zero_prob);
                    }
                }
            }
        }
    }
    // add b to p transition in case of a departure from b arrival state
    for (auto & d_key_queues: queues.queues){
        if (d_key_queues.second.hasKey(KEY_B)){
            double full_prob = d_key_queues.second[KEY_B][capacity];
            d_key_queues.second[KEY_B].setProbability(capacity, 0);
            if (!d_key_queues.second[KEY_B].hasZeroProbability()){
                unblocked += d_key_queues.second[KEY_B].getTotalProbability();
                // there are other non zero probabilities
                d_key_queues.second.addQueueOrInsert(KEY_P, d_key_queues.second[KEY_B]);
                if (full_prob > 0) {
                    d_key_queues.second[KEY_B].emptyQueue();
                    d_key_queues.second[KEY_B].setProbability(capacity, full_prob);
                    
                } else {
                    d_key_queues.second.eraseItem(KEY_B);
                }
            } else if (full_prob > 0){
                d_key_queues.second[KEY_B].setProbability(capacity, full_prob);
            } else {
                d_key_queues.second.eraseItem(KEY_B);
            }
        }
    }
    queues.deleteEmptyQueues();
    assert(checkProbabilityAfterDepartures(queues, was, were_queues, second, downstream_second));
    return unblocked;
}

void getDKeysToRemove(std::vector<Pair> & d_keys_to_remove, std::vector<Pair> & d_keys_not_to_remove){
    // erase all d_key from d_keys_to_remove which have a following key in d_keys_not_to_remove
    for (auto d_iter = d_keys_to_remove.begin(); d_iter != d_keys_to_remove.end(); /*no increment*/){
        bool has_positive_following = std::any_of(d_keys_not_to_remove.begin(), d_keys_not_to_remove.end(), [&](const auto & other_d_key) 
                                             { return other_d_key.first == d_iter->first && other_d_key.second >= d_iter->second;});
        if (has_positive_following){ // remove
            d_iter = d_keys_to_remove.erase(d_iter);
        } else {
            ++d_iter;
        }
    }
}

std::vector<Pair> Queues::cleanUnprobableStates(int step, DepartureChain & departure_chain){
    std::vector<Pair> d_keys_to_remove, d_keys_not_to_remove;
    // first find states
    for (auto & d_iter: queues[step].queues){
        Pair d_key = d_iter.first;
        if (d_key.second <= 0) // do not delete red periods and first green/unblocking state
            continue;
        double prob = 0;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_iter.second.hasKey(a_ind)){
                prob += d_iter.second[a_ind][0];
            }
        }
        prob += departure_chain.getStateProbability(d_key);
        if (prob < almost_zero){ // remove
            d_keys_to_remove.push_back(d_key);
        } else { // check that there is no earlier d_key
            d_keys_not_to_remove.push_back(d_key);
        }
    }
    getDKeysToRemove(d_keys_to_remove, d_keys_not_to_remove);
    // delete from the departure chain and add the following d_keys
    departure_chain.removeStates(d_keys_to_remove);
    // then delete from this and following seconds
    for (int second = step; second < int(queues.size()); ++second){
        for (auto d_key: d_keys_to_remove){
            if (isInMap(queues[second].queues, d_key)){
                queues[second].eraseItem(d_key);
            }
        }
    }
    // clean untransitioned
    untransitioned_queues.removeDepartureStates(d_keys_to_remove); // step is not needed because it is the first available second anyway
    // clean just emptied
    for (auto da_iter = emptied_probs[step].begin(); da_iter != emptied_probs[step].end(); /*no increment*/){
        if (std::find(d_keys_to_remove.begin(), d_keys_to_remove.end(), da_iter->first.first) != d_keys_to_remove.end()){ // found => remove such element
            da_iter = emptied_probs[step].erase(da_iter);
        } else { // not found in to remove -> keep
            ++da_iter;
        }
    }
    return d_keys_to_remove; // is send back to remove in last arrivals
}


int Queues::getTotalNumberOfQueues(){
    int total = 0;
    for (int i = 0; i < int(queues.size()); ++i){
        total += queues[i].getTotalNumberOfQueues();
    }
    return total;
}

std::vector<double> Queues::getDistribution(int second){
    std::vector<double> result;
    int start, end;
    result.resize(capacity + 1, 0.);
    for (auto & d_iter: queues[second].queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_iter.second.hasKey(a_ind)){
                d_iter.second[a_ind].getStartEnd(start, end);
                if (start < end){
                    for (int i = start; i < end; ++i){
                        result[i] += d_iter.second[a_ind][i];
                    }
                }
            }
        }   
    }
    return result;
}

std::vector<double> Queues::getTiltedCumulativeDistribution(int second){
    std::vector<double> result;
    int start, end;
    double total;
    result.resize(capacity + 1, 0.);
    for (auto & d_iter: queues[second].queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_iter.second.hasKey(a_ind)){
                d_iter.second[a_ind].getStartEnd(start, end);
                if (start < end){
                    total = 0;
                    for (int i = end - 1; i >= 0; --i){
                        total += d_iter.second[a_ind][i];
                        result[i] += total;
                    }
                }
            }
        }   
    }
    return result;
}

std::vector<std::vector<double> > Queues::getRightCumulativeDistributions(int start, int end){
    std::vector<std::vector<double>> distributions, right_cimulative;
    double moment;
    int int_moment;
    distributions.resize(end - start);
    right_cimulative.resize(end - start);
    for (int i = 0; i < end - start; ++i){
        distributions[i] = getTiltedCumulativeDistribution(start + i);
        right_cimulative[i].resize(capacity + 1, 0);
    }
    for (int i = 0; i < end - start; ++i){
        for (int j = 0; j <= capacity; ++j){
            moment = std::max(j - 1, 0) * bumper_to_bumper / free_flow_speed + i;
            int_moment = std::min(int(moment), end - start - 1); 
            right_cimulative[i][j] = distributions[int_moment][j];
        }
    }
    return right_cimulative;
}


bool Queues::checkNonZeroProbs(int step, DepartureChain& departure_chain, bool print_all /*=false*/){
    std::map<Pair, double> d_key_probs;
    bool equal = true;
    std::set<Pair> d_keys; // (ordered) set of all d_keys 
    std::map<Pair, double> transitions;
    for (auto const & d_key_queues: queues[step].queues){
        Pair d_key = d_key_queues.first;
        d_keys.insert(d_key);
    }
    for (auto const & d_key: d_keys){
        auto & d_queues = queues[step][d_key];
        double non_zero_prob = getValue(d_key_probs, d_key, 0.);
        double zero_prob = 0;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_queues.hasKey(a_ind)){
                non_zero_prob += d_queues[a_ind].getTotalProbability();
                zero_prob += d_queues[a_ind][0];
                non_zero_prob -= d_queues[a_ind][0];
            }
        }
        if (print_all)
            std::cout << "for " << d_key << " the prob was " << non_zero_prob << "\n";
        if (departure_chain.getLastSecond(d_key) < step) {
            if (isInMap(d_key_probs, d_key)) {
                d_key_probs.erase(d_key);
            }
            transitions = departure_chain.getTransitions(d_key);
            for (auto & other_d_iter: transitions){
                addValue(d_key_probs, other_d_iter.first, other_d_iter.second * non_zero_prob);
                d_keys.insert(other_d_iter.first);
            }
        } else if (d_key.first <= step && d_key.second >= 0){ // not yet departed state => can compare to departure_chain
            if (!almostEqual(non_zero_prob, departure_chain.getStateProbability(d_key), small_eps * 1000)){
                equal = false;
                std::cout << "Non zero prob of " << d_key << " derived from queues is not equal to that of departure chain: " 
                          << non_zero_prob << " != " << departure_chain.getStateProbability(d_key) 
                          << ". The difference is " << non_zero_prob - departure_chain.getStateProbability(d_key) << "\n";
                std::cout << "zero prob is " << zero_prob << "\n";
            } else if (print_all) {
                std::cout << "Non zero prob of " << d_key << " is approx " << non_zero_prob << "\n";
                std::cout << "the difference is " << non_zero_prob - departure_chain.getStateProbability(d_key) << "\n";
                std::cout << "zero prob is " << zero_prob << "\n";
            }
        }
    }
    if (!equal || print_all){
        printQueues(step);
        if (!equal){
            departure_chain.printStates();
        }
    }
    return equal;
}


double Queues::getAverageAndRemove(int second){
    double result = queues[second].getAverageQueue();
    queues[second].eraseAllQueues(); // not that untransitioned queues get deleted themselves
    return result;
}


double Queues::getAverageSquareQueue(int second, int start_from){
    return queues[second].getAverageSquareQueue(start_from);
}


double Queues::getAverageSquareAndRemove(int second, int start_from){
    double result = queues[second].getAverageSquareQueue(start_from);
    queues[second].eraseAllQueues(); // not that untransitioned queues get deleted themselves
    return result;
}
