#include "queues.h"
#include "rollqueues.h"


UntransitionedQueues::UntransitionedQueues(int max_length, int capacity, int begin_with): 
        capacity(capacity), size(max_length), start(0), end(0), extra_in_begin(begin_with){
    queues.resize(size);
}

int UntransitionedQueues::getRightIndex(int second){
    return (second - extra_in_begin + start) % size;
}

bool UntransitionedQueues::isIndexIn(int index){
    return (index >= 0) && (index < size) && 
           (((end >= start) && (index >= start) && (index < end)) || ((end < start) && ((index >= start) || (index < end))));
}

bool UntransitionedQueues::hasSecond(int second){
    return (second >= extra_in_begin) && (second < extra_in_begin + size) && isIndexIn(getRightIndex(second));
}



void UntransitionedQueues::printQueues(int second, bool throw_if_out_of_range /* = true */){
    int index = getRightIndex(second);
    std::cout << "Printing untransitioned queues for second " << second << ":\n";
    if (isIndexIn(index)){
        queues[index].printQueues();
    } else if (throw_if_out_of_range) {
        std::cout << "given second = " << second << ", start = " << start << ", end = " << end << ", size = " << size << "\n";
        throw "printQueues: Incorrect second!\n";
    } else {
        std::cout << second << " <- no info about the queues.\n";
    }
}

void UntransitionedQueues::printTotalProbability(int second){
    int index = getRightIndex(second);
    if (isIndexIn(index)){
        printTotalProbabilityDepArrQueues(queues[index]);
    } else {
        std::cout << "given second = " << second << ", start = " << start << ", end = " << end << ", size = " << size << "\n";
        throw "printTotalProbability: Incorrect second!\n";
    }
}

void UntransitionedQueues::addQueues(int second, DepArrQueues& new_queues){
    int index = getRightIndex(second);
    if (index == end){
        //no such second yet, just push back
        queues[index] = new_queues;
        end = (end + 1) % size;
    } else if (isIndexIn(index)){ // add to existing second
        queues[index].addQueues(new_queues);

    } else {
        std::cout << "given second = " << second << ", start = " << start << ", end = " << end << ", size = " << size << "\n";
        throw "addQueues: Incorrect second!\n";
    }
}


void UntransitionedQueues::addDownstreamPartsOfQueues(int second, int downstream_second, DepArrQueues& new_queues){
    int index = getRightIndex(second);
    if (index == end){
        end = (end + 1) % size;
        queues[index].queues.clear();
    }
    if (isIndexIn(index)){ // add to existing second
        int m_second = minNotAffectedQueueLength(second, downstream_second, capacity);
        for (auto & d_iter: new_queues.queues){
            Pair d_key = d_iter.first;
            for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
                if (d_iter.second.hasKey(a_ind)){
                    int new_start, new_end;
                    d_iter.second[a_ind].getStartEndDownstream(m_second, new_start, new_end);
                    if (new_start < new_end){
                        if (isInMap(queues[index].queues, d_key) && queues[index][d_key].hasKey(a_ind)){
                            queues[index][d_key][a_ind].addDownstreamPartOfQueue(d_iter.second[a_ind], second, downstream_second);   
                        } else {
                            queues[index].emplace(d_key, a_ind, d_iter.second[a_ind], new_start, new_end);   
                        }
                    }
                }
            }
        }
    } else {
        std::cout << "given second = " << second << ", start = " << start << ", end = " << end << ", size = " << size << "\n";
        throw "addDownstreamPartsOfQueues: Incorrect second!\n";
    }
}

void UntransitionedQueues::addMiddleAndUpstreamPartsWithCoef(int second, int downstream_second, DepArrQueues& new_queues, 
                                                             std::map<char, double>& untransitioned_probs, std::map<char, double>& untransitioned_full_probs){
    int index = getRightIndex(second);
    if (index == end){
        end = (end + 1) % size;
        queues[index].queues.clear();
    }
    if (isIndexIn(index)){ // add to existing second
        for (auto & d_iter: new_queues.queues){
            Pair d_key = d_iter.first;
            for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
                if (d_iter.second.hasKey(a_ind)){
                    char a_key = getAKey(a_ind);
                    if (!isInMap(queues[index].queues, d_key) || !queues[index][d_key].hasKey(a_ind)){
                        queues[index][d_key].emplace(a_ind, capacity);   
                    }
                    queues[index][d_key][a_ind].addUpstreamAndMiddleQueueByCoef(new_queues[d_key][a_ind], untransitioned_probs[a_key], 
                                                                                    untransitioned_full_probs[a_key], (a_ind == KEY_A), second, downstream_second);   
                }
            }
        }
    } else {
        std::cout << "given second = " << second << ", start = " << start << ", end = " << end << ", size = " << size << "\n";
        throw "addMiddleAndUpstreamPartsWithCoef: Incorrect second!\n";
    }
}


void UntransitionedQueues::deleteOldQueues(){
    if (end != start){
        queues[start].queues.clear();
        start = (start + 1) % size;
        ++extra_in_begin;
    } else {
        throw "No queues to delete!\n";
    }
}

void UntransitionedQueues::removeMiddleParts(int second, int step){
    int index = getRightIndex(second);
    for (auto & d_key_queues: queues[index].queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues.second.hasKey(a_ind))
                d_key_queues.second[a_ind].removeMiddlePart(second, step);
        }
    }
}


void UntransitionedQueues::removeMiddlePart(int second, Pair d_key, int a_ind, int m_down, int m_up){
    int index = getRightIndex(second);
    if (isInMap(queues[index].queues, d_key) && queues[index][d_key].hasKey(a_ind)){
        queues[index][d_key][a_ind].removeMiddlePartKnowingMs(m_down, m_up);
    }
}


DepArrQueues & UntransitionedQueues::operator[](int second){
    int index = getRightIndex(second);
    if (isIndexIn(index)){
        return queues[index];
    } else {
        std::cout << "given second = " << second << ", start = " << start << ", end = " << end << ", size = " << size << "\n";
        throw "operator[]: Incorrect second!\n";
    }
}


double UntransitionedQueues::getTotalUntransitionedQueue(int second, int downstream_second, char a_key){
    double total = 0;
    int index = getRightIndex(second);
    if (isIndexIn(index)){
        int m_second = minNotAffectedQueueLength(second, downstream_second, capacity);
        int which = (a_key == 'a') ? capacity - 1: capacity;
        for (auto & d_key_iter: queues[index].queues){
            if (d_key_iter.second.hasKey(a_key)){
                int start_queue, end_queue;
                d_key_iter.second[a_key].getStartEnd(start_queue, end_queue);
                for (int i = std::max(start_queue, m_second); i < end_queue && i < which; ++i)
                    total += queues[index][d_key_iter.first][a_key][i];
            }
        }
    } else {
        std::cout << "given second = " << second << ", start = " << start << ", end = " << end << ", size = " << size << "\n";
        throw "getTotalUntransitionedQueue: Incorrect second!\n";
    }
    return total;
}


double UntransitionedQueues::getAlmostFullProbability(int second, char a_key){
    double total = 0;
    int index = getRightIndex(second);
    if (isIndexIn(index)){
        int which = (a_key == 'a') ? capacity - 1: capacity;
        for (auto const& d_key_iter: queues[index].queues){
            if (queues[index][d_key_iter.first].hasKey(a_key)){
                total += queues[index][d_key_iter.first][a_key][which];
            }
        }
    } 
    return total;
}

double UntransitionedQueues::getTotalProbability(){
    double total = 0, prob;
    for (int i = 0; i < size; ++i){
        if (isIndexIn(i)){
            prob = queues[i].getTotalProbability();
            total += prob;
        }
    }
    return total;
}

double UntransitionedQueues::getProbabilityBefore(int second){
    double total = 0;
    int ind = getRightIndex(second);
    int ind_plus = (start > ind) ? ind + size : ind;
    for (int i = start; i < ind_plus; ++i){
        total += queues[i % size].getTotalProbability();
    }
    return total;
}



bool UntransitionedQueues::checkTotalProbability(){
    double total = getTotalProbability();
    if ((total + small_eps < 1) || (total > 1 + small_eps)){
        std::cout << 1 - total << "\n";
        std::cout << start << " " << end << " " << extra_in_begin << "\n";
        std::cout << "see above\n";
        std::cout << "Wrong total untransitioned probability!\n";
        for (int i = start; i < end; ++i){
            printQueues(i + extra_in_begin - start);
        }
        return false;
    }
    return true;
}

bool UntransitionedQueues::checkAllUntransitioned(int second, ArrivalChain& arrival_chain){
    // here untransitioned is everything except transitioned 
    int ind = getRightIndex(second);
    int end_plus = (start > end) ? end + size : end;
    for (int i = ind; i < end_plus; ++i){ 
        int new_second = second + i - ind;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            char a_key = getAKey(a_ind);
            if (arrival_chain.hasState(new_second, a_key)){
                double untransitioned_arrivals = arrival_chain.getUntransitionedProbability(new_second, a_key);
                double untransitioned_queues = 0;
                for (auto & d_key_queues: queues[i % size].queues){
                    if (d_key_queues.second.hasKey(a_ind)){
                        untransitioned_queues += d_key_queues.second[a_ind].getTotalProbability();
                    }
                }
                if (untransitioned_queues > untransitioned_arrivals + eps){
                    printQueues(new_second);
                    arrival_chain.printStates(new_second);
                    std::cout << untransitioned_queues << " " << untransitioned_arrivals << "\n";
                    std::cout << untransitioned_queues - untransitioned_arrivals << "\n";
                    std::cout << new_second << " <- at this second problem\n";
                    std::cout << a_key << " <- at this a_key\n";
                    std::cout << "1 - total probability: " << 1- getTotalProbability() << "\n";
                    std::cout << "Too much probability in untransitioned queues\n";
                    return false;
                }
            }
        }
    }
    return true;
}


bool UntransitionedQueues::checkUntransitioned(int second, ArrivalChain& arrival_chain){
    int ind = getRightIndex(second);
    int end_plus = (start > end) ? end + size : end;    
    int queues_size = extra_in_begin + end_plus - start;
    int arrivals_size = arrival_chain.getSize();
    
    if (queues_size != arrivals_size){
        for (int i = 0; i < arrivals_size; ++i)
            arrival_chain.printStates(i);
        std::cout << queues_size << " " << arrivals_size << "\n";
        std::cout << "Queues size differs from arrivals size.\n";
        return false;
    }
    
    for (int i = ind; i < end_plus; ++i){ 
        int new_second = second + i - ind;
        int m_second = minNotAffectedQueueLength(new_second, second - 1, capacity); // second = step + 1 => second - 1 = step
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            char a_key = getAKey(a_ind);
            if (arrival_chain.hasState(new_second, a_key)){
                double untransitioned_arrivals = arrival_chain.getUnknownTransitionProbability(new_second, a_key);
                double untransitioned_queues = 0;
                for (auto & d_key_queues: queues[i % size].queues){
                    if (d_key_queues.second.hasKey(a_ind)){
                        for (int m = m_second; m < capacity + 1; ++m){
                            untransitioned_queues += d_key_queues.second[a_ind][m];
                        }
                    }
                }
                if (untransitioned_queues > untransitioned_arrivals + small_eps){
                    for (int j = second; j < second - ind + ((end > ind) ? end : end + size); ++j){
                        printQueues(j);
                        arrival_chain.printStates(j);
                    }
                    printQueues(new_second);
                    arrival_chain.printStates(new_second);
                    std::cout << untransitioned_queues << " " << untransitioned_arrivals << "\n";
                    std::cout << untransitioned_queues - untransitioned_arrivals << "\n";
                    std::cout << new_second << " <- at this second problem\n";
                    std::cout << a_key << " <- at this a_key\n";
                    std::cout << "Too much probability in untransitioned queues (in terms of unknown transitions)\n";
                    return false;
                }
            }
        }
    }
    return true;
}

int UntransitionedQueues::getQueuesSize(){
    return (end >= start) ? extra_in_begin + end - start : extra_in_begin + end + size - start;
}


void UntransitionedQueues::removeDepartureStates(const std::vector<Pair>& d_keys_to_remove){
    int end_plus = (start > end) ? end + size : end;
    for (int i = start; i < end_plus; ++i){ 
        for (auto d_key: d_keys_to_remove){
            if (isInMap(queues[i % size].queues, d_key)){
                queues[i % size].eraseItem(d_key);
            }
        }
    }
    
}

void UntransitionedQueues::redirectZeroProbabilityTo(int second, Pair d_zero){
    int ind = getRightIndex(second);
    queues[ind].redirectZeroProbabilityTo(d_zero, capacity);
}


std::map<Pair, double> UntransitionedQueues::getAlmostFullDKeys(int second, char a_key){
    std::map<Pair, double> results;
    double total = 0;
    Pair d_key;
    int which = (a_key == 'a') ? capacity - 1: capacity;
    int ind = getRightIndex(second);
    for (auto & d_key_iter: queues[ind].queues){
        d_key = d_key_iter.first;
        if (d_key_iter.second.hasKey(a_key)){
            addValue(results, d_key, d_key_iter.second[a_key][which]);
            total += d_key_iter.second[a_key][which];
        }
    }
    if (total > almost_zero) {
        for (auto & d_key_iter: queues[ind].queues){
            d_key = d_key_iter.first;
            if (isInMap(results, d_key))
                results[d_key] /= total;
        }
    } else {
        results.clear();
    }
    return results;
}

