#include <iostream>
#include <utility> 
#include "transitions.h"
#include "general.h"


template class Transitions<char>;
// departure pair t, k, where t is time t_g, t_r or t_u, and k is the number of delayed departures so far or -1 for red time
template class Transitions<std::pair<int, int>>;  

template <class type>
void Transitions<type>::addTransition(type key, double value){
    double prev = 0.;
    if (transitions.find(key) != transitions.end())
        prev = transitions[key];
    transitions[key] = value + prev;
}

template<typename type> double Transitions<type>::addDefaultTransition(type key_to, double value){
    // add default transition with such probability that the total transition probability becomes value
    double was_total = sumMapValues(transitions);
    if (value > was_total){
        addTransition(key_to, value - was_total);
    } else if (value + eps < was_total){
        std::cout << key_to << " " << was_total << " " << value << " with diff " << was_total - value << "\n";
        throw "Total transition probabilities are higher than should be!\n";
    }
    return std::max(value - was_total, 0.);
}



std::ostream& operator<<(std::ostream& os, const std::pair<int,int> & pair)
{
    os << "(" << pair.first << ", " << pair.second << ')';
    return os;
}

template <class type>
void Transitions<type>::printTransitions(){
    for (auto const& x : transitions){
        std::cout << x.first << ':' << x.second << " (" << 1 - x.second << ");" << std::endl;
    }
    if (transitions.size() == 0)  // no transitions
        std::cout << "There are no transitions.\n";
}


template <class type>
void Transitions<type>::redirectTransition(type key1, type key2, double prob){
    if (key1 != key2){
        addTransition(key1, -prob);
        addTransition(key2, prob);
        if (transitions[key1] < almost_zero){
            if (transitions[key1] < -small_eps){ // there is a mistake
                printWarning("the probability of the redirected transition is less than the redirected part.");
                std::cout << key1 << key2 << prob << "\n";
                printTransitions();
                throw "Wrong redirected transition!\n";
            }
            transitions.erase(key1);
        }
    }
}

template<class type>
bool Transitions<type>::checkTransitions(double total){
    double real_total = 0;
    for (auto const& x : transitions){
        double prob = x.second;
        real_total += prob;
        if (prob < 0){
            std::cout << x.first << " " << prob << "\n";
//             throw "Negative transition probability\n";
            std::cout << "Negative transition probability\n";
            return false;
        }
    }
    if ((total > real_total + eps) || (real_total > total + eps)){
        printWarning("The total probability of the transitions is wrong! It is " + std::to_string(real_total) + " instead of " + 
                     std::to_string(total) + "!");
//         throw "Wrong transition probabilities!";
        std::cout << "Wrong transition probabilities!";
        return false;
    }
    return true; //total - real_total;
}

template <class type>
std::map<type, double> Transitions<type>::getNormalisedTransitions(double state_probability){
    // Return transitions normalised by state_probability
    std::map<type, double> normalised_transitions;
    if (state_probability > almost_zero){
        for (auto const& x : transitions)
            normalised_transitions[x.first] = x.second / state_probability;
    } // otherwise no need to return any transitions
    return normalised_transitions;
}

template<class type> 
double Transitions<type>::getTotalTransitionProbability(){
    return sumMapValues(transitions);
}


template<typename type> void Transitions<type>::addTransitions(std::map<type, double> & to_add){
    addMap(transitions, to_add);
}

template<typename type> void Transitions<type>::rebalanceTransitions(type default_key, double prob_was, double prob_removed){
    if (transitions.size() > 1){
        double remove_from_default = 0, coef, prob;
        if (prob_was < prob_removed + almost_zero && prob_was > almost_zero){
            std::cout << "Failed to rebalance transitions after " << prob_removed << " probability was removed from " << prob_was << "\n";
            printTransitions();
            throw "Failed to rebalance transitions!\n";
        }
        if (prob_was < prob_removed + almost_zero)
            coef = prob_was / (prob_was - prob_removed);
        else
            coef = 1;
        for (auto & iter : transitions){
            if (iter.first == default_key)
                continue;
            prob = iter.second;
            remove_from_default += prob * (coef - 1);
            iter.second *= coef;
        }
        if (remove_from_default > almost_zero){
            addValue(transitions, default_key, -remove_from_default);
            if (transitions[default_key] < 0){
                std::cout << "too much to remove from the default transition\n";
                printTransitions();
                std::cout << "Failed to rebalance transitions after " << prob_removed << " probability was removed from " << prob_was << "\n";
                throw "Failed to rebalance transitions!\n";
            }
        }
    } // else no need to do anything
}
